import logging
import threading
from cryptoadvance.specter.specter_error import SpecterError
from cryptoadvance.specter.util.requests_tools import failsafe_request_get

logger = logging.getLogger(__name__)


class UtxoScanner:
    """A Command class which simplifies the scanning of UTXOs and the import into a (core+specter) wallet"""

    # 1. pass descriptor (change + recv) and range to rpc.scantxoutset
    #    get list of tx each having deriv-path
    # 2. Update the address-indexes with the max from recv+change (to get fresh addresses and not used once)
    # 3. rpc.getblockhash of height(each tx)
    # 4. rpc.gettxoutproof get a proof that the tx is included in a block, needs tx_id and blockhash for each tx
    #    This will only work on fullnodes (!)

    timeout = (
        300  # 0.001 # As this is async, have longer timeout for all rpc-calls: 5 mins
    )

    def __init__(self, wallet, requests_session, explorer: str = None, only_tor=False):
        if requests_session == None and explorer != None:
            raise SpecterError("Cannot use a blockchainexplorer without a Session!")

        if (
            explorer
            and explorer.endswith(".onion")
            and not requests_session.proxies["https"]
        ):
            raise SpecterError(
                "Cannot Query an Onion Explorer without a working Tor Setup! Check your Tor Setup or choose a non Tor Explorer."
            )
        self.wallet = wallet
        self.explorer = explorer
        self.requests_session = requests_session
        if self.explorer:
            self.explorer = self.explorer.rstrip("/")
        self.only_tor = only_tor
        self.error_msgs = []

    def execute(self, asyncc=False):
        if asyncc:
            t = threading.Thread(
                target=self._execute,
            )
            t.start()
        else:
            self._execute()

    def _execute(self):
        self.args = self.get_scantxoutset_args()
        # get something like:
        # [{'txid': '917a4d55...', 'vout': 1, 'scriptPubKey': '0014c313...5b19317',
        #   'desc': "wpkh([e3b947d9/84'/1'/0'/1/0]0244290480...ab2f94d940a7a)#gey9w279",
        #   'amount': 0.02119209, 'height': 1939628
        # }]
        self.unspents = self.wallet.rpc.scantxoutset(*self.args, timeout=self.timeout)[
            "unspents"
        ]
        logger.info(f"Found {len(self.unspents)} utxo")
        self.adjust_keypools()
        # Endgoal: calling importprunedfunds with the raw-transaction and a proof that the tx was included in a block
        self.add_blockhashes_to_txs_in_unspents()
        self.add_proofs_to_txs_in_unspents()
        self.add_rawtxs_to_txs_in_unspents()
        self.missing = [tx for tx in self.unspents if tx["raw"] is None]
        self.existing = [tx for tx in self.unspents if tx["raw"] is not None]
        # Now we're ready to import at least the existing txs
        logger.info(f"Importing {len(self.existing)} utxos to core-wallet")
        self.wallet.rpc.multi(
            [("importprunedfunds", tx["raw"], tx["proof"]) for tx in self.existing],
            timeout=self.timeout,
        )
        if len(self.missing) == 0:
            logger.info(f"no more missing utxos. Rescan completed successfully!")
            return self.execute_post_processing()
        # let's continue with the missing ones via an explorer
        if not self.check_explorer_working():
            logger.error(f"Completed unsuccessfully!")
            return self.execute_post_processing()
        self.add_proofs_to_txs_in_missing_via_exlorer()
        self.add_rawtxs_to_txs_in_missing_via_exlorer()
        self.existing_via_explorer = [
            tx for tx in self.missing if tx["raw"] is not None
        ]
        self.missing_even_after_explorer = [
            tx for tx in self.missing if tx["raw"] is None
        ]
        # Now importing the missing ones
        logger.info(
            f"Importing {len(self.existing_via_explorer)} utxos to core-wallet (found via Explorer)"
        )
        self.wallet.rpc.multi(
            [
                ("importprunedfunds", tx["raw"], tx["proof"])
                for tx in self.existing_via_explorer
            ],
            timeout=self.timeout,
        )
        if len(self.missing_even_after_explorer) != 0:
            logger.error(
                f"Even after explorer {len(self.missing_even_after_explorer)} txs could not be resolved"
            )
            logger.error(
                f"Giving up, here is a list of the still missing TXIDs: {[ tx['txid'] for tx in self.missing_even_after_explorer]}"
            )
            self.error_msgs.append(
                f"Some TXs could not be imported: {[ tx['txid'] for tx in self.missing_even_after_explorer]}"
            )
        return self.execute_post_processing()

    def execute_post_processing(self):
        self.wallet.fetch_transactions()
        self.wallet.check_addresses()

    def get_scantxoutset_args(self):
        return [
            "start",
            [
                {
                    "desc": self.wallet.recv_descriptor,
                    "range": max(self.wallet.keypool, 1000),
                },
                {
                    "desc": self.wallet.change_descriptor,
                    "range": max(self.wallet.change_keypool, 1000),
                },
            ],
        ]

    def adjust_keypools(self):
        """check the unspent for max indexes of recv+change addresses and adjusts the index of the wallet (if needed)
        saves the wallet if changed.
        """
        # check derivation indexes in found unspents (last 2 indexes in [brackets])
        derivations = [
            #                      purpose / account / cointype / change / index
            # {... 'desc': "wpkh([e3b947d9/84'/1'/0'/1/0]02442...7a)#gey9w279"
            tx["desc"].split("[")[1].split("]")[0].split("/")[-2:]
            for tx in self.unspents
        ]
        # we get a list of change / index parts of the derivations
        # [['1', '0'],['1', '1'],['0', '2']] represents 2 change and one recv-address (index 2)

        # get the maximum index for both address-types:
        max_recv = max([-1] + [int(der[1]) for der in derivations if der[0] == "0"])
        max_change = max([-1] + [int(der[1]) for der in derivations if der[0] == "1"])

        updated = False
        if max_recv >= self.wallet.address_index:
            # skip to max_recv
            self.wallet.address_index = max_recv
            logger.info(f"Adjusted address_index of {self.wallet} to {max_change}")
            # get next
            self.wallet.getnewaddress(change=False, save=False)
            updated = True
        if max_change >= self.wallet.change_index:
            # skip to max_change
            self.wallet.change_index = max_change
            logger.info(f"Adjusted change_index of {self.wallet} to {max_change}")
            # get next
            self.wallet.getnewaddress(change=True, save=False)
            updated = True
        # save only if needed
        if updated:
            self.wallet.save_to_file()

    def add_blockhashes_to_txs_in_unspents(self):
        """includes the blockhash of the block each tx in unspents has been mined into"""
        res = self.wallet.rpc.multi(
            [("getblockhash", tx["height"]) for tx in self.unspents],
            timeout=self.timeout,
        )
        block_hashes = [r["result"] for r in res]
        for i, tx in enumerate(self.unspents):
            # each tx in the unspents get the hash of the block it has been included in
            tx["blockhash"] = block_hashes[i]

    def add_proofs_to_txs_in_unspents(self):
        """includes the proof that the txs in unspents has been mined into the block"""
        res = self.wallet.rpc.multi(
            [("gettxoutproof", [tx["txid"]], tx["blockhash"]) for tx in self.unspents],
            timeout=self.timeout,
        )
        proofs = [r["result"] for r in res]
        for i, tx in enumerate(self.unspents):
            tx["proof"] = proofs[i]

    def add_rawtxs_to_txs_in_unspents(self):
        """includes the raw tx to the txs in unspents"""
        res = self.wallet.rpc.multi(
            [
                ("getrawtransaction", tx["txid"], False, tx["blockhash"])
                for tx in self.unspents
            ],
            timeout=self.timeout,
        )
        raws = [r["result"] for r in res]
        for i, tx in enumerate(self.unspents):
            tx["raw"] = raws[i]

    def check_explorer_working(self):
        """returns a boolean if the explorer link is usable. Will log.error and fill self.error_msgs if not"""
        if self.explorer is None:
            logger.error(
                f"Not all txs could be resolved and can't use explorer to get the rest"
            )
            self.error_msgs.append(
                "Not all txs could be resolved and can't use explorer to get the rest"
            )
            return False
        if not self.explorer.startswith("http"):
            logger.error(f"explorer seem to have an invalid url: {self.explorer}")
            self.error_msgs.append(
                f"explorer seem to have an invalid url: {self.explorer}"
            )
            return False
        try:
            request = failsafe_request_get(
                self.requests_session, f"{self.explorer}", parse_json=False
            )
        except SpecterError:
            return False
        if not request.ok:
            logger.error(f"explorer seem to have an invalid url: {self.explorer}")
            self.error_msgs.append(
                f"explorer seem to have an invalid url: {self.explorer}"
            )
            return False
        return True

    def add_proofs_to_txs_in_missing_via_exlorer(self):
        proofs = [
            self.requests_session.get(
                f"{self.explorer}/api/tx/{tx['txid']}/merkleblock-proof"
            ).text
            for tx in self.missing
        ]
        for i, tx in enumerate(self.missing):
            tx["proof"] = proofs[i]

    def add_rawtxs_to_txs_in_missing_via_exlorer(self):
        raws = [
            self.requests_session.get(f"{self.explorer}/api/tx/{tx['txid']}/hex").text
            for tx in self.missing
        ]
        for i, tx in enumerate(self.missing):
            tx["raw"] = raws[i]



import copy
import json
import logging
import os
import platform
import random
import secrets
import signal
import threading
import time
import traceback
import zipfile
from io import BytesIO
from sys import exit
from urllib.parse import urlparse

import requests
from requests.exceptions import ConnectionError
from stem.control import Controller
from urllib3.exceptions import NewConnectionError

from cryptoadvance.specter.devices.device_types import DeviceTypes

from .helpers import clean_psbt, deep_update, is_liquid, is_testnet, get_asset_label
from .internal_node import InternalNode
from .liquid.rpc import LiquidRPC
from .managers.config_manager import ConfigManager
from .managers.node_manager import NodeManager
from .managers.otp_manager import OtpManager
from .managers.user_manager import UserManager
from .managers.wallet_manager import WalletManager
from .node import Node
from .persistence import read_json_file, write_json_file, write_node
from .process_controller.bitcoind_controller import BitcoindPlainController
from .rpc import (
    BitcoinRPC,
    RpcError,
    get_default_datadir,
)
from .managers.service_manager import ServiceManager
from .services.service import devstatus_alpha, devstatus_beta, devstatus_prod
from .specter_error import ExtProcTimeoutException, SpecterError
from .tor_daemon import TorDaemonController
from .user import User
from .util.checker import Checker
from .util.version import VersionChecker
from .util.price_providers import update_price
from .util.setup_states import SETUP_STATES
from .util.tor import get_tor_daemon_suffix

logger = logging.getLogger(__name__)


class Specter:
    """A central Object mostly holding app-settings"""

    # use this lock for all fs operations
    lock = threading.Lock()
    _default_asset = None

    def __init__(
        self,
        data_folder="./data",
        config={},
        internal_bitcoind_version="",
        checker_threads=True,
    ):
        if data_folder.startswith("~"):
            data_folder = os.path.expanduser(data_folder)
        data_folder = os.path.abspath(data_folder)

        # creating folders if they don't exist
        if not os.path.isdir(data_folder):
            os.makedirs(data_folder)

        self.data_folder = data_folder

        self.user_manager = UserManager(
            self
        )  # has to come before calling VersionChecker()

        # version checker
        # checks for new versions once per hour
        self.version = VersionChecker(specter=self)
        if checker_threads:
            self.version.start()

        self._config_manager = ConfigManager(self.data_folder, config)

        self.internal_bitcoind_version = internal_bitcoind_version

        # Migrating from Specter 1.3.1 and lower (prior to the node manager)
        self.migrate_old_node_format()

        self.node_manager = NodeManager(
            proxy_url=self.proxy_url,
            only_tor=self.only_tor,
            active_node=self.active_node_alias,
            bitcoind_path=self.bitcoind_path,
            internal_bitcoind_version=internal_bitcoind_version,
            data_folder=os.path.join(self.data_folder, "nodes"),
        )

        self.torbrowser_path = os.path.join(
            self.data_folder, f"tor-binaries/tor{get_tor_daemon_suffix()}"
        )

        self._tor_daemon = None

        self.setup_status = {
            "stage": "start",
            "bitcoind": {
                "stage_progress": -1,
                "stage": "",
                "error": "",
            },
            "torbrowser": {
                "stage_progress": -1,
                "stage": "",
                "error": "",
            },
        }
        # health check: loads config, tests rpc
        # also loads and checks wallets for all users
        try:
            self.check(check_all=True)

            if os.path.isfile(self.torbrowser_path):
                self.tor_daemon.start_tor_daemon()
        except Exception as e:
            logger.error(e)

        self.update_tor_controller()
        self.checker = Checker(lambda: self.check(check_all=True), desc="health")
        if checker_threads:
            self.checker.start()
        self.price_checker = Checker(
            lambda: update_price(self, self.user), desc="price"
        )
        if self.price_check and self.price_provider and checker_threads:
            self.price_checker.start()

        if threading.current_thread() is threading.main_thread():
            # Python signal handlers are always executed in the main Python thread and only the main thread is allowed to set a new signal handler.
            # This is for CTRL-C --> SIGINT
            signal.signal(signal.SIGINT, self.cleanup_on_exit)
            # This is for kill $pid --> SIGTERM
            signal.signal(signal.SIGTERM, self.cleanup_on_exit)

    def cleanup_on_exit(self, signum=0, frame=0):
        if self._tor_daemon:
            logger.info("Specter exit cleanup: Stopping Tor daemon")
            self._tor_daemon.stop_tor_daemon()

        for node in self.node_manager.nodes.values():
            if not node.external_node:
                node.stop()

        logger.info("Closing Specter after cleanup")
        # For some reason we need to explicitely exit here. Otherwise it will hang
        exit(0)

    def check(self, user=None, check_all=False):
        """
        Checks and updates everything for a particular user:
        - config if changed
        - rpc including check if it's connected
        - node info
        - wallet manager
        - device manager
        """
        # check if config file have changed
        self.check_config()

        self.node.update_rpc()

        # if rpc is not available
        # do checks more often, once in 20 seconds
        if self.rpc is None or self.node.info.get("initialblockdownload", True):
            period = 20
        else:
            period = 600
        if hasattr(self, "checker") and self.checker.period != period:
            self.checker.period = period

        if not check_all:
            # find proper user
            user = self.user_manager.get_user(user)
            user.check()
        else:
            for u in self.user_manager.users:
                u.check()

    @property
    def node(self):
        try:
            return self.node_manager.active_node
        except SpecterError as e:
            logger.error("SpecterError while accessing active_node")
            logger.exception(e)
            self.update_active_node(list(self.node_manager.nodes.values())[0].alias)
            return self.node_manager.active_node

    @property
    def default_node(self):
        return self.node_manager.default_node()

    @property
    def rpc(self):
        return self.node.rpc

    @property
    def utxorescanwallet(self):
        return self.node.utxorescanwallet

    @utxorescanwallet.setter
    def utxorescanwallet(self, value):
        self.node.utxorescanwallet = value

    @property
    def config(self):
        """A convenience property simply redirecting to the config_manager"""
        return self.config_manager.data

    def check_blockheight(self):
        if self.node.check_blockheight():
            self.check(check_all=True)

    def get_user_folder_id(self, user=None):
        """
        Returns the suffix for the user wallets and devices.
        User can be either a flask_login user or a string.
        """
        user = self.user_manager.get_user(user)
        if not user.is_admin:
            return "_" + user.id
        return ""

    def check_config(self):
        """
        Updates config if file config have changed.
        Priority (low to high):
        - existing / default config
        - file config from config.json
        - arg_config passed in constructor
        """
        self.config_manager.check_config()

    def delete_user(self, user):
        if user not in self.user_manager.users:
            return
        user = self.user_manager.get_user(user)
        user.wallet_manager.delete(self)
        user.device_manager.delete(self)
        self.user_manager.delete_user(user)

    # mark
    @property
    def bitcoin_datadir(self):
        return self.node.datadir

    # mark
    def _save(self):
        write_json_file(self.config, self.config_fname, lock=self.lock)

    @property
    def config_fname(self):
        return os.path.join(self.data_folder, "config.json")

    # mark
    def update_active_node(self, node_alias):
        """update the current active node to use"""
        self.config_manager.update_active_node(node_alias)
        self.node_manager.switch_node(node_alias)
        self.check()

    def update_setup_status(self, software_name, stage):
        self.setup_status[software_name]["error"] = ""
        if stage in SETUP_STATES:
            self.setup_status[software_name]["stage"] = SETUP_STATES[stage].get(
                software_name, stage
            )
        else:
            self.setup_status[software_name]["stage"] = stage
        self.setup_status[software_name]["stage_progress"] = 0

    def update_setup_download_progress(self, software_name, progress):
        self.setup_status[software_name]["error"] = ""
        self.setup_status[software_name]["stage_progress"] = progress

    def update_setup_error(self, software_name, error):
        self.setup_status[software_name]["error"] = error
        self.setup_status[software_name]["stage_progress"] = -1

    def reset_setup(self, software_name):
        self.setup_status[software_name]["error"] = ""
        self.setup_status[software_name]["stage"] = ""
        self.setup_status[software_name]["stage_progress"] = -1

    def get_setup_status(self, software_name):
        if software_name == "bitcoind":
            installed = os.path.isfile(self.bitcoind_path)
        elif software_name == "torbrowser":
            installed = os.path.isfile(self.torbrowser_path)
        else:
            installed = False

        return {"installed": installed, **self.setup_status[software_name]}

    # mark
    def update_auth(self, method, rate_limit, registration_link_timeout):
        """simply persisting the current auth-choice"""
        self.config_manager.update_auth(method, rate_limit, registration_link_timeout)

    # mark
    def update_explorer(self, explorer_id, explorer_data, user):
        """update the block explorers urls"""
        self.config_manager.update_explorer(
            explorer_id, explorer_data, user, self.chain
        )

    # mark
    def update_fee_estimator(self, fee_estimator, custom_url, user):
        """update the fee estimator option and its url if custom"""
        self.config_manager.update_fee_estimator(fee_estimator, custom_url, user)

    # mark
    def update_tor_type(self, tor_type, user):
        """update the Tor proxy url"""
        if tor_type == "builtin":
            self.update_proxy_url("socks5h://localhost:9050", user)
            self.update_tor_control_port("", user)
        self.config_manager.update_tor_type(tor_type, user)

    # mark
    def update_proxy_url(self, proxy_url, user):
        """update the Tor proxy url"""
        self.config_manager.update_proxy_url(proxy_url, user)

    # mark
    def toggle_tor_status(self):
        """toggle the Tor status"""
        self.config_manager.toggle_tor_status()

    # mark
    def update_only_tor(self, only_tor, user):
        """switch whatever to use Tor for all calls"""
        self.config_manager.update_only_tor(only_tor, user)

    # mark
    def update_tor_control_port(self, tor_control_port, user):
        """set the control port of the tor daemon"""
        if self.config_manager.update_tor_control_port:
            self.update_tor_controller()

    # mark
    def generate_torrc_password(self, overwrite=False):
        self.config_manager.generate_torrc_password(overwrite)

    def update_tor_controller(self):
        if "torrc_password" not in self.config:
            # Will be missing if the user did not go through the built-in Tor setup
            self.generate_torrc_password()
        try:
            tor_control_address = urlparse(self.proxy_url).netloc.split(":")[0]
            if tor_control_address == "localhost":
                tor_control_address = "127.0.0.1"
            self._tor_controller = Controller.from_port(
                address=tor_control_address,
                port=int(self.tor_control_port) if self.tor_control_port else "default",
            )
            self._tor_controller.authenticate(
                password=self.config.get("torrc_password", "")
            )
        except Exception as e:
            logger.warning(f"Failed to connect to Tor control port. Error: {e}")
            self._tor_controller = None

    @property
    def tor_daemon(self):
        if os.path.isfile(self.torbrowser_path) and os.path.join(
            self.data_folder, "torrc"
        ):
            if not self._tor_daemon:
                self._tor_daemon = TorDaemonController(
                    tor_daemon_path=self.torbrowser_path,
                    tor_config_path=os.path.join(self.data_folder, "torrc"),
                )
            return self._tor_daemon
        raise SpecterError(
            "Tor daemon files missing. Make sure Tor is installed within Specter"
        )

    def is_tor_dameon_running(self):
        return self._tor_daemon and self._tor_daemon.is_running()

    @property
    def tor_controller(self):
        if self._tor_controller:
            return self._tor_controller
        self.update_tor_controller()
        if self._tor_controller:
            return self._tor_controller
        raise SpecterError(
            "Failed to connect to the Tor daemon. Make sure ControlPort is properly configured."
        )

    # mark
    def update_hwi_bridge_url(self, url, user):
        """update the hwi bridge url to use"""
        self.config_manager.update_hwi_bridge_url(url, user)

    # mark
    def update_unit(self, unit, user):
        self.config_manager.update_unit(unit, user)

    # mark
    def update_hide_sensitive_info(self, hide_sensitive_info_bool, user):
        self.config_manager.update_hide_sensitive_info(hide_sensitive_info_bool, user)

    # mark
    def update_price_check_setting(self, price_check_bool, user):
        self.config_manager.update_price_check_setting(price_check_bool, user)

    # mark
    def update_price_provider(self, price_provider, user):
        self.config_manager.update_price_provider(price_provider, user)

    # mark needs User-Type injection
    def update_weight_unit(self, weight_unit, user):
        self.config_manager.update_weight_unit(weight_unit, user)

    # mark needs User-Type injection
    def update_alt_rate(self, alt_rate, user):
        self.config_manager.update_alt_rate(alt_rate, user)

    # mark
    def update_alt_symbol(self, alt_symbol, user):
        self.config_manager.update_alt_symbol(alt_symbol, user)

    def update_services(self, services):
        """takes a list of service_names which should be activated"""
        self.config["services"] = services
        self._save()

    def update_merkleproof_settings(self, validate_bool):
        if validate_bool is True and self.info.get("pruned") is True:
            validate_bool = False
            logger.warning("Cannot enable merkleproof setting on pruned node.")
        self.config_manager.update_merkleproof_settings(validate_bool)

    def combine(self, psbt_arr):
        # backward compatibility with current Core psbt parser
        psbt_arr = [clean_psbt(psbt) for psbt in psbt_arr]
        final_psbt = self.rpc.combinepsbt(psbt_arr)
        return final_psbt

    def finalize(self, psbt):
        final_psbt = self.rpc.finalizepsbt(psbt)
        return final_psbt

    def broadcast(self, raw):
        res = self.rpc.sendrawtransaction(raw)
        return res

    def estimatesmartfee(self, blocks):
        res = self.rpc.estimatesmartfee(blocks)
        if "feerate" not in res and self.is_liquid:
            return 0.000001
        return res

    @property
    def bitcoind_path(self):
        bitcoind_path = os.path.join(self.data_folder, "bitcoin-binaries/bin/bitcoind")

        if platform.system() == "Windows":
            bitcoind_path += ".exe"
        return bitcoind_path

    @property
    def info(self):
        return self.node.info

    @property
    def network_info(self):
        return self.node.network_info

    @property
    def bitcoin_core_version(self):
        return self.node.bitcoin_core_version

    @property
    def bitcoin_core_version_raw(self):
        return self.node.bitcoin_core_version_raw

    @property
    def taproot_support(self):
        return self.node.taproot_support

    @property
    def chain(self):
        return self.node.chain

    @property
    def network_parameters(self):
        return self.node.network_parameters

    @property
    def is_testnet(self):
        return self.node.is_testnet

    @property
    def is_liquid(self):
        return is_liquid(self.chain)

    @property
    def user_config(self):
        return self.config if self.user.is_admin else self.user.config

    @property
    def active_node_alias(self):
        return self.user_config.get("active_node_alias", "default")

    @property
    def explorer(self):
        return self.user_config.get("explorers", {}).get(self.chain, "")

    @property
    def explorer_id(self):
        return self.user_config.get("explorer_id", {}).get(self.chain, "CUSTOM")

    @property
    def asset_labels(self):
        user_assets = self.user_config.get("asset_labels", {}).get(self.chain, {})
        node_assets = self.node.asset_labels
        asset_labels = {}
        deep_update(asset_labels, node_assets)
        deep_update(asset_labels, user_assets)
        return asset_labels

    @property
    def default_asset(self):
        """returns hash of LBTC"""
        if self._default_asset is None:
            for asset, lbl in self.asset_labels.items():
                if lbl in ["LBTC", "tLBTC"]:
                    self._default_asset = asset
                    return asset
        return self._default_asset

    def asset_label(self, asset):
        if asset == "":
            return ""
        return get_asset_label(asset, known_assets=self.asset_labels)

    def update_asset_label(self, asset, label):
        if asset == self.default_asset:
            raise SpecterError("LBTC should stay LBTC")
        self.config_manager.update_asset_label(asset, label, self.chain, self.user)

    @property
    def fee_estimator(self):
        return self.user_config.get("fee_estimator", "bitcoin_core")

    @property
    def tor_type(self):
        return self.user_config.get("tor_type", "builtin")

    @property
    def proxy_url(self):
        return self.user_config.get("proxy_url", "socks5h://localhost:9050")

    @property
    def only_tor(self):
        return self.user_config.get("only_tor", False)

    @property
    def tor_control_port(self):
        return self.user_config.get("tor_control_port", "")

    @property
    def hwi_bridge_url(self):
        return self.user_config.get("hwi_bridge_url", "")

    @property
    def unit(self):
        return self.user_config.get("unit", "btc")

    @property
    def price_check(self):
        return self.user_config.get("price_check", False)

    @property
    def price_provider(self):
        return self.user_config.get("price_provider", False)

    @property
    def weight_unit(self):
        return self.user_config.get("weight_unit", "oz")

    @property
    def alt_rate(self):
        return self.user_config.get("alt_rate", 1)

    @property
    def alt_symbol(self):
        return self.user_config.get("alt_symbol", "BTC")

    @property
    def admin(self):
        for u in self.user_manager.users:
            if u.is_admin:
                return u

    @property
    def user(self):
        return self.user_manager.user

    @property
    def config_manager(self):
        if not hasattr(self, "_config_manager"):
            self._config_manager = ConfigManager(self.data_folder)
        return self._config_manager

    @property
    def device_manager(self):
        return self.user.device_manager

    @property
    def wallet_manager(self):
        return self.user.wallet_manager

    @property
    def otp_manager(self):
        if not hasattr(self, "_otp_manager"):
            self._otp_manager = OtpManager(self.data_folder)
        return self._otp_manager

    @property
    def hide_sensitive_info(self):
        return self.user_config.get("hide_sensitive_info", False)

    @property
    def autohide_sensitive_info_timeout(self):
        return self.user_config.get("autohide_sensitive_info_timeout_minutes", None)

    @property
    def autologout_timeout(self):
        return self.user_config.get("autologout_timeout_hours", 4)

    def requests_session(self, force_tor=False):
        requests_session = requests.Session()
        if self.only_tor or force_tor:
            proxy_url = self.proxy_url
            proxy_parsed_url = urlparse(self.proxy_url)
            proxy_url = proxy_parsed_url._replace(
                netloc="{}:{}@{}".format(
                    str(random.randint(10000, 0x7FFFFFFF)),
                    "random",
                    proxy_parsed_url.netloc,
                )
            ).geturl()
            requests_session.proxies["http"] = proxy_url
            requests_session.proxies["https"] = proxy_url
        return requests_session

    def specter_backup_file(self):
        memory_file = BytesIO()
        with zipfile.ZipFile(memory_file, "w") as zf:
            if self.wallet_manager:
                for wallet in self.wallet_manager.wallets.values():
                    data = zipfile.ZipInfo("{}.json".format(wallet.alias))
                    data.date_time = time.localtime(time.time())[:6]
                    data.compress_type = zipfile.ZIP_DEFLATED
                    zf.writestr(
                        "wallets/{}.json".format(wallet.alias),
                        json.dumps(wallet.to_json(for_export=True)),
                    )
            if self.device_manager:
                for device in self.device_manager.devices.values():
                    data = zipfile.ZipInfo("{}.json".format(device.alias))
                    data.date_time = time.localtime(time.time())[:6]
                    data.compress_type = zipfile.ZIP_DEFLATED
                    device = device.json
                    # Exporting the bitcoincore hot wallet as watchonly
                    if device["type"] == DeviceTypes.BITCOINCORE:
                        device["type"] = DeviceTypes.BITCOINCORE_WATCHONLY
                    zf.writestr(
                        "devices/{}.json".format(device["alias"]), json.dumps(device)
                    )
        memory_file.seek(0)
        return memory_file

    # Migrating RPC nodes from Specter 1.3.1 and lower (prior to the node manager)
    def migrate_old_node_format(self):
        if not os.path.isdir(os.path.join(self.data_folder, "nodes")):
            os.mkdir(os.path.join(self.data_folder, "nodes"))
        old_rpc = self.config.get("rpc", None)
        old_internal_rpc = self.config.get("internal_node", None)
        if old_internal_rpc and os.path.isfile(self.bitcoind_path):
            internal_node = InternalNode(
                "Specter Bitcoin",
                "specter_bitcoin",
                old_internal_rpc.get("autodetect", False),
                old_internal_rpc.get("datadir", get_default_datadir()),
                old_internal_rpc.get("user", ""),
                old_internal_rpc.get("password", ""),
                old_internal_rpc.get("port", 8332),
                old_internal_rpc.get("host", "localhost"),
                old_internal_rpc.get("protocol", "http"),
                os.path.join(
                    os.path.join(self.data_folder, "nodes"), "specter_bitcoin.json"
                ),
                self,
                self.bitcoind_path,
                "mainnet",
                "0.20.1",
            )
            logger.info(f"persisting {internal_node} in migrate_old_node_format")
            write_node(
                internal_node,
                os.path.join(
                    os.path.join(self.data_folder, "nodes"), "specter_bitcoin.json"
                ),
            )
            del self.config["internal_node"]
            if not old_rpc or not old_rpc.get("external_node", True):
                self.config_manager.update_active_node("specter_bitcoin")

        if old_rpc:
            node = Node(
                "Bitcoin Core",
                "default",
                old_rpc.get("autodetect", True),
                old_rpc.get("datadir", get_default_datadir()),
                old_rpc.get("user", ""),
                old_rpc.get("password", ""),
                old_rpc.get("port", None),
                old_rpc.get("host", "localhost"),
                old_rpc.get("protocol", "http"),
                True,
                os.path.join(os.path.join(self.data_folder, "nodes"), "default.json"),
                "BTC",
                self,
            )
            logger.info(f"persisting {node} in migrate_old_node_format")
            write_node(
                node,
                os.path.join(os.path.join(self.data_folder, "nodes"), "default.json"),
            )
            del self.config["rpc"]
        self._save()


class SpecterConfiguration:
    """An abstract class which only holds functionality relevant for storage of information mostly
    deferring to ConfigManager.
    Do not put logic in here, which is not directly relevant for the config.json.
    Do not deal with the config.json directly but implement that in the ConfigManager
    """

    pass
    # ToDo: move all the methods above here.



from . import DeviceTypes
from .hwi_device import HWIDevice
from .hwi.jade import JadeClient
from .hwi.jade import enumerate as jade_enumerate
from ..helpers import is_liquid


class Jade(HWIDevice):
    device_type = DeviceTypes.JADE
    name = "Jade"
    icon = "jade_icon.svg"

    supports_hwi_toggle_passphrase = False
    supports_hwi_multisig_display_address = True
    liquid_support = True

    @classmethod
    def get_client(cls, *args, **kwargs):
        return JadeClient(*args, **kwargs)

    @classmethod
    def enumerate(cls, *args, **kwargs):
        return jade_enumerate(*args, **kwargs)

    def has_key_types(self, wallet_type, network="main"):
        if wallet_type == "multisig" and is_liquid(network):
            return False
        return super().has_key_types(wallet_type, network)

    def no_key_found_reason(self, wallet_type, network="main"):
        if wallet_type == "multisig" and is_liquid(network):
            return "Jade does not support multisig wallets on Liquid."
        return super().no_key_found_reason(wallet_type, network)



""" Provides an API abstracting writes (and reads) on anything
    happening on the ~/.specter-folder. Useful as we can then
    call the call-back-method.
"""

import csv
import json
import logging
import os
import shutil
import threading


from flask import current_app as app

from .specter_error import SpecterError
from .services.callbacks import specter_persistence_callback
from .util.shell import run_shell

logger = logging.getLogger(__name__)

fslock = threading.Lock()
pclock = threading.Lock()


def read_json_file(path):
    """read_json_file from the .specter-directory. Don't use it for
    something else"""
    with fslock:
        bkp = path + ".bkp"
        # try reading file
        try:
            with open(path, "r") as f:
                content = json.load(f)

        # if failed - try reading from the backup
        except Exception as e:
            logger.error(
                f"Exception {e} while reading file {path}. Reading from backup"
            )
            # if no backup exists - raise
            if not os.path.isfile(bkp):
                raise e
            # try loading from backup
            with open(bkp, "r") as f:
                content = json.load(f)
            # recover from backup
            if os.path.isfile(path):
                os.remove(path)
            os.rename(bkp, path)
            logger.error(f"failed to open {path}, recovered from backup.")
        return content


def _delete_folder(path):
    """Internal method which won't trigger the callback"""
    with fslock:
        shutil.rmtree(path, ignore_errors=True)  # might not exist


def _write_json_file(content, path, lock=None):
    """Internal method which won't trigger the callback"""
    if lock is None:
        lock = fslock
    with lock:
        # backup file
        bkp = path + ".bkp"
        # check if file exists
        if os.path.isfile(path):
            # check if backup exists
            if os.path.isfile(bkp):
                # remove backup file
                os.remove(bkp)
            # move file to backup
            os.rename(path, bkp)
        try:
            with open(path, "w") as f:
                json.dump(content, f, indent=4)

            # check if write was sucessful
            with open(path, "r") as f:
                c = json.load(f)

        # if not - move back backup
        except Exception as e:
            # remove damaged file
            if os.path.isfile(path):
                os.remove(path)
            shutil.copyfile(bkp, path)
            logger.exception(e)
            raise SpecterError(
                f"Error:{path} could not be saved. The old version has been restored. Check the logs for details. This is probably a bug."
            )


def write_json_file(content, path, lock=None):
    _write_json_file(content, path, lock)
    storage_callback(path=path)


def delete_files(paths):
    """deletes multiple files and calls storage callback once"""
    for path in paths:
        try:
            os.remove(path)
            storage_callback(mode="delete", path=path)
        except FileNotFoundError:
            pass


def delete_file(path):
    delete_files([path])


def write_devices(devices_json):
    """interpret a json as a list of devices and write them in the devices subfolder inside the specter-folder"""
    for device_json in devices_json:
        path = os.path.join(
            app.specter.device_manager.data_folder, "%s.json" % device_json["alias"]
        )
        _write_json_file(device_json, path)
        storage_callback(path=path)


def write_wallet(wallet_json):
    """interpret a json as wallet and writes it in the wallets subfolder inside the specter-folder.
    overwrites it if existing.
    """
    fpath = os.path.join(
        app.specter.wallet_manager.working_folder, "%s.json" % wallet_json["alias"]
    )
    _write_json_file(wallet_json, fpath)
    storage_callback(path=fpath)


def write_device(device, fullpath):
    _write_json_file(device.json, fullpath)
    storage_callback(path=fullpath)


def write_node(node, fullpath):
    _write_json_file(node.json, fullpath)
    storage_callback(path=fullpath)


def delete_folder(path):
    _delete_folder(path)
    storage_callback(mode="delete", path=path)


def delete_folders(paths):
    for path in paths:
        _delete_folder(path)
        storage_callback(mode="delete", path=path)


def _write_csv(fname, objs, cls=dict):
    columns = []
    # if it's a custom class
    if hasattr(cls, "columns"):
        columns = cls.columns
    # if it's just a dict
    elif len(objs) > 0:
        columns = objs[0].keys()
    with fslock:
        with open(fname, mode="w") as csv_file:
            writer = csv.DictWriter(csv_file, fieldnames=columns, extrasaction="ignore")
            writer.writeheader()
            for obj in objs:
                writer.writerow(obj)


def write_csv(fname, objs, cls=dict):
    _write_csv(fname, objs, cls)
    storage_callback(path=fname)


def read_csv(fname, cls=dict, *args):
    with fslock:
        with open(fname, mode="r") as csv_file:
            csv_reader = csv.DictReader(csv_file)
            return [cls(*args, **row) for row in csv_reader]


def storage_callback_function(cmd_list):
    """This is executing the callback-script, either directly or via threading"""
    with pclock:
        logger.debug(f"Executing {cmd_list}")
        result = run_shell(cmd_list)
        if result["code"] != 0:
            logger.error("callback failed ")
            logger.error("stderr: {}".format(result["err"]))
            logger.error("stdout: {}".format(result["out"]))
        else:
            logger.info("Successfully executed {}".format(" ".join(cmd_list)))
            logger.debug("result: {}".format(result))


def storage_callback(mode="write", path=None):
    """Call this whenever anything in the .specter directory changes. Be aware that we might store node-data in the specter-folder"""
    # Might be usefull to figure out why the callback has been triggered:
    # traceback.print_stack()
    # logger.debug(f"Storage Callback called mode {mode} with path {path}")

    # First, call extensions which want to get informed.
    # This is working synchronously!
    try:
        app.specter.service_manager.execute_ext_callbacks(
            specter_persistence_callback, path=path, mode=mode
        )
    except AttributeError as e:
        # chicken-egg poroblem:
        if str(e).endswith("object has no attribute 'specter'"):
            pass
        else:
            raise e
    except RuntimeError as e:
        if str(e).startswith("Working outside of application context"):
            # not yet (?!) supported
            pass
        else:
            raise e

    # Now maybe we have async scripts?
    if os.getenv("SPECTER_PERSISTENCE_CALLBACK_ASYNC"):
        cmd_list = os.getenv("SPECTER_PERSISTENCE_CALLBACK_ASYNC").split(" ")
        cmd_list.append(mode)
        cmd_list.append(path)
        t = threading.Thread(
            target=storage_callback_function,
            args=(cmd_list,),
        )
        t.start()
    # Or maybe even synchronous scripts?
    elif os.getenv("SPECTER_PERSISTENCE_CALLBACK"):
        cmd_list = os.getenv("SPECTER_PERSISTENCE_CALLBACK").split(" ")
        cmd_list.append(mode)
        cmd_list.append(path)
        storage_callback_function(cmd_list)



#!/usr/bin/env python3
# Copyright (c) 2015-2018 The Bitcoin Core developers
# Distributed under the MIT software license, see the accompanying
# file COPYING or http://www.opensource.org/licenses/mit-license.php.

from argparse import ArgumentParser
from base64 import urlsafe_b64encode
from binascii import hexlify
from getpass import getpass
from os import urandom

import hmac


def generate_salt(size):
    """Create size byte hex salt"""
    return hexlify(urandom(size)).decode()


def generate_password():
    """Create 32 byte b64 password"""
    return urlsafe_b64encode(urandom(32)).decode("utf-8")


def password_to_hmac(salt, password):
    m = hmac.new(bytearray(salt, "utf-8"), bytearray(password, "utf-8"), "SHA256")
    return m.hexdigest()


def main():
    parser = ArgumentParser(description="Create login credentials for a JSON-RPC user")
    parser.add_argument("username", help="the username for authentication")
    parser.add_argument(
        "password",
        help='leave empty to generate a random password or specify "-" to prompt for password',
        nargs="?",
    )
    args = parser.parse_args()

    if not args.password:
        args.password = generate_password()
    elif args.password == "-":
        args.password = getpass()

    # Create 16 byte hex salt
    salt = generate_salt(16)
    password_hmac = password_to_hmac(salt, args.password)

    print("String to be appended to bitcoin.conf:")
    print("rpcauth={0}:{1}${2}".format(args.username, salt, password_hmac))
    print("Your password:\n{0}".format(args.password))


if __name__ == "__main__":
    main()



import subprocess
import sys
import logging
import re
import threading
import time
import os
from urllib.error import HTTPError
import requests
from requests.exceptions import ConnectionError
from urllib3.exceptions import NewConnectionError
import importlib_metadata

from cryptoadvance.specter.specter_error import SpecterError

logger = logging.getLogger(__name__)


class VersionChecker:
    def __init__(self, name="cryptoadvance.specter", specter=None):
        self.name = name

        self.current = self._get_current_version()
        self.latest = "unknown"
        self.upgrade = False
        self.running = False
        self.specter = specter

    def start(self):
        if not self.running:
            self.running = True
            self.thread = threading.Thread(target=self.loop)
            self.thread.daemon = True
            self.thread.start()

    def stop(self):
        logger.info("version checker stopped.")
        self.running = False

    @property
    def info(self):
        info = {"current": self.current, "latest": self.latest, "upgrade": self.upgrade}
        return info

    @property
    def installation_type(self):
        """app or pip"""
        if getattr(sys, "frozen", False):
            return "app"
        else:
            return "pip"

    def loop(self, dt=3600):
        """Checks for updates once per hour"""
        while self.running:
            self.current, self.latest, self.upgrade = self.get_version_info()
            logger.info(
                f"version checked, install_type {self.installation_type} curr: {self.current} latest: {self.latest} ==> upgrade: {self.upgrade}"
            )
            time.sleep(dt)

    def _get_current_version(self):
        current = "unknown"
        if self.installation_type == "app":
            current = VersionChecker._version_txt_content()
        else:
            current = importlib_metadata.version(self.name)
            if current == None:
                current = "custom"
        if current == "vx.y.z-get-replaced-by-release-script":
            current = "custom"
        return current

    def _get_binary_version(self):
        """
        Get binary version: current, latest.
        Fails if version.txt is not present.
        Returns latest = "unknown" if fetch failed.
        """
        current = self._get_current_version()
        latest = "unknown"
        if self.name != "cryptoadvance.specter":
            logger.warning(
                "We're checking here for a different binary than specter-desktop. We're hopefully in a pytest"
            )
        latest = self._get_latest_version_from_github()
        return current, latest

    def _get_pip_version(self):
        """
        returns current, latest
        """
        current = self._get_current_version()
        latest = "unknown"
        try:
            if self.specter:
                requests_session = self.specter.requests_session(force_tor=False)
            else:
                requests_session = requests.Session()

            releases = (
                requests_session.get(f"https://pypi.org/pypi/{self.name}/json")
                .json()["releases"]
                .keys()
            )
            releases = list(releases)
            for i in range(-1, 0 - len(releases), -1):
                # for some stupid reason, rc-versions are BEFORE the real versions
                if not releases[i][:-1].endswith("rc"):
                    latest = releases[i]
                    break
        except (
            HTTPError,
            ConnectionError,
            ConnectionRefusedError,
            NewConnectionError,
        ) as e:
            logger.error(f"{e} while checking for new pypi version")
        except Exception as e:
            logger.exception(e)
            latest = "unknown"
        return current, latest

    def get_version_info(self):
        """
        Returns a triple of the current version
        of the pip-package cryptoadvance.specter and
        the latest version and whether you should upgrade.
        """
        # check if we have version.txt file
        # this is the case for binaries
        current = "unknown"
        latest = "unknown"
        # check binary version
        if self.installation_type == "app":
            current, latest = self._get_binary_version()
        else:
            current, latest = self._get_pip_version()

        # check that both current and latest versions match the pattern
        # vA.B.C or just A.B.C
        # For vA.B.C-preX or something like that we don't show notification
        if re.search(r"v?([\d+]).([\d+]).([\d+])$", current):
            if re.search(r"v?([\d+]).([\d+]).([\d+])$", latest):
                return (
                    current,
                    latest,
                    # check without leading v so v1.2.3 = 1.2.3
                    latest.replace("v", "") != current.replace("v", ""),
                )
        # if current version is not A.B.C - stop periodic checks
        else:
            self.stop()
        return current, latest, False

    def _get_latest_version_from_github(self):
        try:
            if self.specter:
                requests_session = self.specter.requests_session(force_tor=False)
            else:
                requests_session = requests.Session()
            releases = (
                requests_session.get(f"https://pypi.org/pypi/{self.name}/json")
                .json()["releases"]
                .keys()
            )
            latest = list(releases)[-1]
        except Exception as e:
            logger.exception(e)
            latest = "unknown"
        return latest

    @classmethod
    def _version_txt_content(cls):
        version_file = "version.txt"
        if getattr(sys, "frozen", False):
            version_file = os.path.join(sys._MEIPASS, "version.txt")
        with open(version_file) as f:
            return f.read().strip()


def compare(version1: str, version2: str) -> int:
    """Compares two version strings like v1.5.1 and v1.6.0 and returns
    * 1 : version2 is bigger that version1
    * -1 : version1 is bigger than version2
    * 0 : both are the same
    This is not supporting semver and it doesn't take any postfix (-pre5)
    into account and is therefore a naive implementation
    """
    version1 = _parse_version(version1)
    version2 = _parse_version(version2)

    if version1["postfix"] != "" or version2["postfix"] != "":
        raise SpecterError(
            f"Cannot compare if either version has a postfix : {version1} and {version2}"
        )
    if version1["major"] > version2["major"]:
        return -1
    elif version1["major"] < version2["major"]:
        return 1
    if version1["minor"] > version2["minor"]:
        return -1
    elif version1["minor"] < version2["minor"]:
        return 1
    if version1["patch"] > version2["patch"]:
        return -1
    elif version1["patch"] < version2["patch"]:
        return 1
    return 0


def _parse_version(version: str) -> dict:
    """Parses version-strings like v1.5.6-pre5 and returns a dict"""
    if version[0] != "v":
        raise SpecterError(f"version {version} does not have a preceding 'v'")
    version = version[1:]
    version_ar = version.split(".")
    if len(version_ar) != 3:
        raise SpecterError(f"version {version} does not have 3 separated digits")
    postfix = ""
    if "-" in version_ar[2]:
        postfix = version_ar[2].split("-")[1]
        version_ar[2] = version_ar[2].split("-")[0]
    return {
        "major": int(version_ar[0]),
        "minor": int(version_ar[1]),
        "patch": int(version_ar[2]),
        "postfix": postfix,
    }



import socket
import logging


logger = logging.getLogger("jade.tcp")


#
# Low-level Serial-via-TCP backend interface to Jade
# Calls to send and receive bytes over the interface.
# Intended for use via JadeInterface wrapper.
#
# Either:
#  a) use via JadeInterface.create_serial() (see JadeInterface)
# (recommended)
# or:
#  b) use JadeTCPImpl() directly, and call connect() before
#     using, and disconnect() when finished,
# (caveat cranium)
#
class JadeTCPImpl:
    PROTOCOL_PREFIX = "tcp:"

    @classmethod
    def isSupportedDevice(cls, device):
        return device is not None and device.startswith(cls.PROTOCOL_PREFIX)

    def __init__(self, device):
        assert self.isSupportedDevice(device)
        self.device = device
        self.tcp_sock = None

    def connect(self):
        assert self.isSupportedDevice(self.device)
        assert self.tcp_sock is None

        logger.info("Connecting to {}".format(self.device))
        self.tcp_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        url = self.device[len(self.PROTOCOL_PREFIX) :].split(":")
        self.tcp_sock.connect((url[0], int(url[1])))
        assert self.tcp_sock is not None

        self.tcp_sock.__enter__()
        logger.info("Connected")

    def disconnect(self):
        assert self.tcp_sock is not None
        self.tcp_sock.__exit__()

        # Reset state
        self.tcp_sock = None

    def write(self, bytes_):
        assert self.tcp_sock is not None
        return self.tcp_sock.send(bytes_)

    def read(self, n):
        assert self.tcp_sock is not None
        return self.tcp_sock.recv(n)



from hwilib.psbt import PSBT
from hwilib.common import Chain
import hwilib.commands as hwi_commands
from .helpers import locked
from .util.xpub import convert_xpub_prefix
from .util.json_rpc import JSONRPC
from .util.descriptor import AddChecksum
import threading
from .devices import __all__ as device_classes
from contextlib import contextmanager
from embit import bip32
from embit.liquid import networks
import logging
import bitbox02
from typing import Callable
from flask import current_app as app
from .helpers import (
    deep_update,
    hwi_get_config,
    save_hwi_bridge_config,
    is_testnet,
    is_liquid,
)
from hwilib.devices.bitbox02 import Bitbox02Client
from hwilib.devices.trezorlib.transport import get_transport
from .devices.hwi.specter_diy import SpecterClient
from .devices.hwi.jade import JadeClient

logger = logging.getLogger(__name__)

hwi_classes = [cls for cls in device_classes if cls.hwi_support]

# use this lock for all hwi operations
hwilock = threading.Lock()


class BitBox02NoiseConfig(bitbox02.util.BitBoxAppNoiseConfig):
    def show_pairing(self, code: str, device_response: Callable[[], bool]) -> bool:
        config = hwi_get_config(app.specter)
        config["bitbox02_pairing_code"] = code
        save_hwi_bridge_config(app.specter, config)
        if not device_response():
            config["bitbox02_pairing_code"] = ""
            save_hwi_bridge_config(app.specter, config)
            return False
        config["bitbox02_pairing_code"] = ""
        save_hwi_bridge_config(app.specter, config)
        return True


def get_device_class(device_type):
    for cls in hwi_classes:
        if cls.device_type == device_type:
            return cls


class HWIBridge(JSONRPC):
    """
    A class that represents HWI JSON-RPC methods.

    All methods of this class are callable over JSON-RPC, except _underscored.
    """

    def __init__(self):
        self.exposed_rpc = {
            "enumerate": self.enumerate,
            "detect_device": self.detect_device,
            "toggle_passphrase": self.toggle_passphrase,
            "prompt_pin": self.prompt_pin,
            "send_pin": self.send_pin,
            "extract_xpub": self.extract_xpub,
            "extract_xpubs": self.extract_xpubs,
            "display_address": self.display_address,
            "sign_tx": self.sign_tx,
            "sign_message": self.sign_message,
            "extract_master_blinding_key": self.extract_master_blinding_key,
            "bitbox02_pairing": self.bitbox02_pairing,
        }
        # Running enumerate after beginning an interaction with a specific device
        # crashes python or make HWI misbehave. For now we just get all connected
        # devices once per session and save them.
        logger.info("Initializing HWI...")  # to explain user why it takes so long
        self.enumerate()

    @locked(hwilock)
    def enumerate(self, passphrase="", chain=""):
        """
        Returns a list of all connected devices (dicts).
        Standard HWI enumerate() command + Specter.
        """
        devices = []
        # going through all device classes
        for devcls in hwi_classes:
            try:
                # calling device-specific enumerate
                if passphrase:
                    devs = devcls.enumerate(passphrase)
                # not sure if it will handle passphrase correctly
                # so remove it if None
                else:
                    devs = devcls.enumerate()
                # extracting fingerprint info
                for dev in devs:
                    # we can't get fingerprint if device is locked
                    if "needs_pin_sent" in dev and dev["needs_pin_sent"]:
                        continue
                    # we can't get fingerprint if passphrase is not provided
                    if (
                        "needs_passphrase_sent" in dev
                        and dev["needs_passphrase_sent"]
                        and not passphrase
                    ):
                        continue
                    client = None
                    try:
                        client = devcls.get_client(dev["path"], passphrase)
                        if isinstance(client, Bitbox02Client):
                            client.set_noise_config(BitBox02NoiseConfig())
                        dev["fingerprint"] = client.get_master_fingerprint().hex()
                    finally:
                        if client is not None:
                            client.close()
                devices += devs
            except Exception as e:
                logger.warning(f"enumerate failed: {e}")

        self.devices = devices
        return self.devices

    def detect_device(
        self, device_type=None, path=None, fingerprint=None, rescan_devices=False
    ):
        """
        Returns a hardware wallet details
        with specific fingerprint/ path/ type
        or None if not connected.
        If found multiple devices return only one.
        """
        if rescan_devices:
            self.enumerate()
        res = []

        if device_type is not None:
            res = [
                dev
                for dev in self.devices
                if dev["type"].lower() == device_type.lower()
            ]
        if fingerprint is not None:
            res = [
                dev
                for dev in self.devices
                if dev["fingerprint"].lower() == fingerprint.lower()
            ]
        if path is not None:
            res = [dev for dev in self.devices if dev["path"] == path]
        if len(res) > 0:
            return res[0]

    @locked(hwilock)
    def toggle_passphrase(self, device_type=None, path=None, passphrase="", chain=""):
        if device_type == "keepkey" or device_type == "trezor":
            with self._get_client(
                device_type=device_type, path=path, passphrase=passphrase, chain=chain
            ) as client:
                return hwi_commands.toggle_passphrase(client)
        else:
            raise Exception(
                "Invalid HWI device type %s, toggle_passphrase is only supported for Trezor and Keepkey devices"
                % device_type
            )

    @locked(hwilock)
    def prompt_pin(self, device_type=None, path=None, passphrase="", chain=""):
        if device_type == "keepkey" or device_type == "trezor":
            # The device will randomize its pin entry matrix on the device
            #   but the corresponding digits in the receiving UI always map
            #   to:
            #       7 8 9
            #       4 5 6
            #       1 2 3
            with self._get_client(
                device_type=device_type, path=path, passphrase=passphrase, chain=chain
            ) as client:
                return hwi_commands.prompt_pin(client)
        else:
            raise Exception(
                "Invalid HWI device type %s, prompt_pin is only supported for Trezor and Keepkey devices"
                % device_type
            )

    @locked(hwilock)
    def send_pin(self, pin="", device_type=None, path=None, passphrase="", chain=""):
        if device_type == "keepkey" or device_type == "trezor":
            if pin == "":
                raise Exception("Must enter a non-empty PIN")
            with self._get_client(
                device_type=device_type, path=path, passphrase=passphrase, chain=chain
            ) as client:
                logger.debug(f"client is : {client}")
                return hwi_commands.send_pin(client, pin)
        else:
            raise Exception(
                "Invalid HWI device type %s, send_pin is only supported for Trezor and Keepkey devices"
                % device_type
            )

    @locked(hwilock)
    def extract_xpubs(
        self,
        account=0,
        device_type=None,
        path=None,
        fingerprint=None,
        passphrase="",
        chain="",
    ):
        with self._get_client(
            device_type=device_type,
            fingerprint=fingerprint,
            path=path,
            passphrase=passphrase,
            chain=chain,
        ) as client:
            xpubs = self._extract_xpubs_from_client(client, account)
        return xpubs

    @locked(hwilock)
    def extract_xpub(
        self,
        derivation=None,
        device_type=None,
        path=None,
        fingerprint=None,
        passphrase="",
        chain="",
    ):
        with self._get_client(
            device_type=device_type,
            fingerprint=fingerprint,
            path=path,
            passphrase=passphrase,
            chain=chain,
        ) as client:
            # Client will be configured for testnet if our Specter instance is
            #   currently connected to testnet. This will prevent us from
            #   getting mainnet xpubs unless we set is_testnet here:
            der = bip32.parse_path(derivation)
            client.chain = (
                Chain.TEST if len(der) > 2 and der[1] == 0x80000001 else Chain.MAIN
            )

            network = networks.get_network(
                "main" if client.chain == Chain.MAIN else "test"
            )

            master_fpr = client.get_master_fingerprint().hex()

            try:
                xpub = client.get_pubkey_at_path(derivation).to_string()
                slip132_prefix = bip32.detect_version(
                    derivation, default="xpub", network=network
                )
                xpub = convert_xpub_prefix(xpub, slip132_prefix)
                if derivation == "m":
                    return "[{}]{}\n".format(master_fpr, xpub)
                return "[{}/{}]{}\n".format(master_fpr, derivation.split("m/")[1], xpub)
            except Exception as e:
                logger.warning(
                    f"Failed to import Nested Segwit singlesig mainnet key. Error: {e}"
                )

    @locked(hwilock)
    def display_address(
        self,
        descriptor="",
        xpubs_descriptor="",
        device_type=None,
        path=None,
        fingerprint=None,
        passphrase="",
        chain="",
    ):
        if descriptor == "" and xpubs_descriptor == "":
            raise Exception("Descriptor must not be empty")

        with self._get_client(
            device_type=device_type,
            fingerprint=fingerprint,
            path=path,
            passphrase=passphrase,
            chain=chain,
        ) as client:
            if xpubs_descriptor:
                status = hwi_commands.displayaddress(client, desc=xpubs_descriptor)
            else:
                status = hwi_commands.displayaddress(client, desc=descriptor)
            if "error" in status:
                raise Exception(status["error"])
            elif "address" in status:
                return status["address"]
            else:
                raise Exception("Failed to validate address on device: Unknown Error")

    @locked(hwilock)
    def sign_tx(
        self,
        psbt="",
        device_type=None,
        path=None,
        fingerprint=None,
        passphrase="",
        chain="",
    ):
        if psbt == "":
            raise Exception("PSBT must not be empty")
        with self._get_client(
            device_type=device_type,
            fingerprint=fingerprint,
            path=path,
            passphrase=passphrase,
            chain=chain,
        ) as client:
            if is_liquid(chain):
                if not hasattr(client, "sign_pset"):
                    raise Exception("Device can't sign liquid transaction")
                return client.sign_pset(psbt)
            status = hwi_commands.signtx(client, psbt)
            if "error" in status:
                raise Exception(status["error"])
            elif "psbt" in status:
                return status["psbt"]
            else:
                raise Exception("Failed to sign transaction with device: Unknown Error")

    @locked(hwilock)
    def sign_message(
        self,
        message="",
        derivation_path="m",
        device_type=None,
        path=None,
        fingerprint=None,
        passphrase="",
        chain="",
    ):
        if message == "":
            raise Exception("Message must not be empty")
        with self._get_client(
            device_type=device_type,
            fingerprint=fingerprint,
            path=path,
            passphrase=passphrase,
            chain=chain,
        ) as client:
            status = hwi_commands.signmessage(client, message, derivation_path)
            if "error" in status:
                raise Exception(status["error"])
            elif "signature" in status:
                return status["signature"]
            else:
                raise Exception("Failed to sign message with device: Unknown Error")

    @locked(hwilock)
    def extract_master_blinding_key(
        self,
        device_type=None,
        path=None,
        fingerprint=None,
        passphrase="",
        chain="",
    ):
        with self._get_client(
            device_type=device_type,
            fingerprint=fingerprint,
            path=path,
            passphrase=passphrase,
            chain=chain,
        ) as client:
            try:
                return client.get_master_blinding_key()
            except Exception as e:
                logger.warning(
                    f"Failed to get the master blinding key from the device. Error: {e}"
                )

    def bitbox02_pairing(self, chain=""):
        config = hwi_get_config(app.specter)
        return {"code": config.get("bitbox02_pairing_code", "")}

    ######################## HWI Utils ########################
    @contextmanager
    def _get_client(
        self, device_type=None, path=None, fingerprint=None, passphrase="", chain=""
    ):
        """
        Returns a hardware wallet class instance
        with specific fingerprint or/and path
        or raises a not found error if not connected.
        If found multiple devices return only one.
        """
        # We do not use fingerprint in most cases since if the device is a trezor
        # or a keepkey and passphrase is enabled but empty (an empty string like '')
        # The device will not return the fingerprint properly.
        device = self.detect_device(
            device_type=device_type, fingerprint=fingerprint, path=path
        )
        if not device:
            raise Exception(
                "The device could not be found. Please check it is properly connected and try again"
            )
        devcls = get_device_class(device["type"])
        if devcls:
            client = devcls.get_client(device["path"], passphrase)
        if not client:
            raise Exception(
                "The device was identified but could not be reached.  Please check it is properly connected and try again"
            )
        try:
            if is_liquid(chain):
                client.chain = Chain.TEST if is_testnet(chain) else Chain.MAIN
            else:
                client.chain = Chain.argparse(chain)
            # hack for Jade
            if type(client) is JadeClient:
                if is_liquid(chain):
                    client.set_liquid_network(chain)
                elif chain == "signet":
                    client.chain = Chain.TEST
            yield client
        finally:
            client.close()

    def _extract_xpubs_from_client(self, client, account=0):
        try:
            xpubs = ""
            # Client will be configured for testnet if our Specter instance is
            #   currently connected to testnet. This will prevent us from
            #   getting mainnet xpubs unless we set is_testnet here:
            client.chain = Chain.MAIN

            master_fpr = client.get_master_fingerprint().hex()

            # HWI calls to client.get_pubkey_at_path() return "xpub"-prefixed xpubs
            # regardless of derivation path. Update to match SLIP-0132 prefixes.
            # See:
            #   https://github.com/satoshilabs/slips/blob/master/slip-0132.md

            # Extract nested Segwit
            try:
                xpub = client.get_pubkey_at_path(
                    "m/49h/0h/{}h".format(account)
                ).to_string()
                ypub = convert_xpub_prefix(xpub, b"\x04\x9d\x7c\xb2")
                xpubs += "[{}/49'/0'/{}']{}\n".format(master_fpr, account, ypub)
            except Exception as e:
                logger.warning(
                    f"Failed to import Nested Segwit singlesig mainnet key. Error {e}"
                )

            try:
                # native Segwit
                xpub = client.get_pubkey_at_path(
                    "m/84h/0h/{}h".format(account)
                ).to_string()
                zpub = convert_xpub_prefix(xpub, b"\x04\xb2\x47\x46")
                xpubs += "[{}/84'/0'/{}']{}\n".format(master_fpr, account, zpub)
            except Exception as e:
                logger.warning(
                    f"Failed to import native Segwit singlesig mainnet key: {e}"
                )

            try:
                # Multisig nested Segwit
                xpub = client.get_pubkey_at_path(
                    "m/48h/0h/{}h/1h".format(account)
                ).to_string()
                Ypub = convert_xpub_prefix(xpub, b"\x02\x95\xb4\x3f")
                xpubs += "[{}/48'/0'/{}'/1']{}\n".format(master_fpr, account, Ypub)
            except Exception as e:
                logger.warning(
                    f"Failed to import Nested Segwit multisig mainnet key: {e}"
                )

            try:
                # Multisig native Segwit
                xpub = client.get_pubkey_at_path(
                    "m/48h/0h/{}h/2h".format(account)
                ).to_string()
                Zpub = convert_xpub_prefix(xpub, b"\x02\xaa\x7e\xd3")
                xpubs += "[{}/48'/0'/{}'/2']{}\n".format(master_fpr, account, Zpub)
            except Exception as e:
                logger.warning(
                    f"Failed to import native Segwit multisig mainnet key {e}"
                )

            # And testnet
            client.chain = Chain.TEST

            try:
                # Testnet nested Segwit
                xpub = client.get_pubkey_at_path(
                    "m/49h/1h/{}h".format(account)
                ).to_string()
                upub = convert_xpub_prefix(xpub, b"\x04\x4a\x52\x62")
                xpubs += "[{}/49'/1'/{}']{}\n".format(master_fpr, account, upub)
            except Exception as e:
                logger.warning(
                    f"Failed to import Nested Segwit singlesig testnet key: {e}"
                )

            try:
                # Testnet native Segwit
                xpub = client.get_pubkey_at_path(
                    "m/84h/1h/{}h".format(account)
                ).to_string()
                vpub = convert_xpub_prefix(xpub, b"\x04\x5f\x1c\xf6")
                xpubs += "[{}/84'/1'/{}']{}\n".format(master_fpr, account, vpub)
            except Exception as e:
                logger.warning(
                    f"Failed to import native Segwit singlesig testnet key: {e}"
                )

            try:
                # Testnet multisig nested Segwit
                xpub = client.get_pubkey_at_path(
                    "m/48h/1h/{}h/1h".format(account)
                ).to_string()
                Upub = convert_xpub_prefix(xpub, b"\x02\x42\x89\xef")
                xpubs += "[{}/48'/1'/{}'/1']{}\n".format(master_fpr, account, Upub)
            except Exception as e:
                logger.warning(
                    f"Failed to import Nested Segwit multisigsig testnet key: {e}"
                )

            try:
                # Testnet multisig native Segwit
                xpub = client.get_pubkey_at_path(
                    "m/48h/1h/{}h/2h".format(account)
                ).to_string()
                Vpub = convert_xpub_prefix(xpub, b"\x02\x57\x54\x83")
                xpubs += "[{}/48'/1'/{}'/2']{}\n".format(master_fpr, account, Vpub)
            except Exception as e:
                logger.warning(
                    f"Failed to import native Segwit multisig testnet key: {e}"
                )

            # Do proper cleanup otherwise have to reconnect device to access again
            client.close()
        except Exception as e:
            if client:
                client.close()
            raise e
        return xpubs



from ..txlist import *
from embit.liquid.transaction import LTransaction, TxOutWitness, unblind
from embit.liquid.pset import PSET
from embit.liquid import slip77
from embit.hashes import tagged_hash
from embit.ec import PrivateKey
from .util.pset import SpecterLTx, get_value, get_asset, SpecterPSET
from io import BytesIO
from embit.psbt import read_string


class LTxItem(TxItem):
    TransactionCls = LTransaction
    columns = [
        "txid",  # str, txid in hex
        "blockhash",  # str, blockhash, None if not confirmed
        "blockheight",  # int, blockheight, None if not confirmed
        "time",  # int (timestamp in seconds), time received
        "bip125-replaceable",  # str ("yes" / "no"), whatever RBF is enabled for the transaction
        "conflicts",  # rbf conflicts, list of txids
        "vsize",
        "category",
        "address",
        "amount",
        "asset",
        "ismine",
    ]
    type_converter = [
        str,
        str,
        int,
        int,
        str,
        parse_arr,
        int,
        str,
        parse_arr,
        parse_arr,
        parse_arr,
        bool,
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # unblind what's blinded and remove
        # will fill automatically
        self.vsize
        self._unblind()

    def _unblind(self):
        if not self.descriptor.is_blinded:
            return

        b = self.tx

        mbpk = self.descriptor.blinding_key.key
        net = self.network

        values = [0 for out in b.vout]
        assets = [b"\xFF" * 32 for out in b.vout]
        datas = []
        # search for datas encoded in rangeproofs
        for i, out in enumerate(b.vout):
            # unblinded
            if isinstance(out.value, int):
                values[i] = out.value
                assets[i] = out.asset
                continue

            pk = slip77.blinding_key(mbpk, out.script_pubkey)
            try:
                res = out.unblind(pk.secret, message_length=1000)
                value, asset, vbf, abf, extra, *_ = res
                if len(extra.rstrip(b"\x00")) > 0:
                    datas.append(extra)
                values[i] = value
                assets[i] = asset
            except Exception as e:
                logger.warn(e)  # TODO: remove, it's ok
                pass

        # to calculate blinding seed
        tx = PSET(b)
        seed = tagged_hash("liquid/blinding_seed", mbpk.secret)
        txseed = tx.txseed(seed)
        pubkeys = {}

        for extra in datas:
            s = BytesIO(extra)
            while True:
                k = read_string(s)
                if len(k) == 0:
                    break
                v = read_string(s)
                if k[0] == 1 and len(k) == 5:
                    idx = int.from_bytes(k[1:], "little")
                    pubkeys[idx] = v
                elif k == b"\x01\x00":
                    txseed = v

        for i, out in enumerate(b.vout):
            if out.witness.range_proof.is_empty:
                continue
            if i in pubkeys and len(pubkeys[i]) in [33, 65]:
                nonce = tagged_hash(
                    "liquid/range_proof", txseed + i.to_bytes(4, "little")
                )
                if out.ecdh_pubkey == PrivateKey(nonce).sec():
                    try:
                        res = unblind(
                            pubkeys[i],
                            nonce,
                            out.witness.range_proof.data,
                            out.value,
                            out.asset,
                            out.script_pubkey,
                        )
                        value, asset, vbf, abf, extra, min_value, max_value = res
                        assets[i] = asset
                        values[i] = value
                    except Exception as e:
                        logger.warn(f"Failed at unblinding output {i}: {e}")
                else:
                    logger.warn(f"Failed at unblinding: {e}")

        for i, out in enumerate(b.vout):
            out.asset = assets[i]
            out.value = values[i]
            out.witness = TxOutWitness()

    @property
    def vsize(self):
        if self.get("vsize"):
            return self["vsize"]
        tx = self.tx
        txsize = len(tx.serialize())
        # tx size - flag - marker - witness
        non_witness_size = (
            txsize
            - 2
            - sum([len(inp.witness.serialize()) for inp in tx.vin])
            - sum([len(out.witness.serialize()) for out in tx.vout])
        )
        witness_size = txsize - non_witness_size
        weight = non_witness_size * 4 + witness_size
        vsize = math.ceil(weight / 4)
        return vsize

    def __dict__(self):
        return {
            "txid": self["txid"],
            "blockhash": self["blockhash"],
            "blockheight": self["blockheight"],
            "time": self["time"],
            "conflicts": self["conflicts"],
            "bip125-replaceable": self["bip125-replaceable"],
            "vsize": self["vsize"],
            "category": self["category"],
            "address": self["address"],
            "amount": self["amount"],
            "asset": self["asset"],
            "ismine": self["ismine"],
        }


class LTxList(TxList):
    ItemCls = LTxItem
    PSBTCls = SpecterPSET
    counter = 0

    def _get_psbt(self, raw_tx):
        psbt = self.PSBTCls.from_transaction(raw_tx, self.descriptor, self.network)
        psbt.psbt.version = 2
        # fill derivation paths etc
        updated = self.rpc.walletprocesspsbt(str(psbt), False).get("psbt", None)
        if updated:
            psbt.update(updated)
        return psbt

    def _update_destinations(self, tx, outs):
        # remove dummy and fee
        outs = [out for out in outs if out.get("address") not in ["Fee", "DUMMY"]]
        # process the rest
        addresses = [out.get("address", "Unknown") for out in outs]
        amounts = [out.get("float_amount", 0) for out in outs]
        assets = [out.get("asset", "ff" * 32) for out in outs]
        if len(addresses) == 1:
            addresses = addresses[0]
            amounts = amounts[0]
            assets = assets[0]
        tx["address"] = addresses
        tx["amount"] = amounts
        tx["asset"] = assets

    def decoderawtransaction(self, tx: Union[LTransaction, str, bytes]):
        return SpecterLTx(self, tx).to_dict()



""" The migrationManager handles all things necessary migrating from one specter version to another.
    It uses patterns from DB-migration-frameworks like flask-migrate and the like.
    The idea is to keep adding migration scripts to the SpecterMigrator and they are executed exactly
    once. In order to keep track of that, the SpecterMigrator uses a MigDataManager which stores
    events and migration_executions in a file called migration_data.json.
    * Events are purely informational and only for troubleshooting purposes.
    * migration_executions simply keeps track of which scrips has been executed.

"""
import inspect
import logging
import os
from datetime import datetime
from importlib import import_module
from inspect import isclass
from pathlib import Path
from pkgutil import iter_modules

from cryptoadvance import specter

from ..managers.genericdata_manager import GenericDataManager
from ..specter_error import SpecterError
from .version import VersionChecker, compare

logger = logging.getLogger(__name__)


class SpecterMigration:
    """All Migrations should derive from this class. In order to create one, have a look at
    the example in the migrations directory.
    """

    version = "custom"  # the version this has been introduced

    def __init__(self, specter_migrator):
        self.specter_migrator = specter_migrator
        self.data_folder = specter_migrator.data_folder

    def should_execute(self):
        """You can override this to prevent execution for specific cases
        The default is that the version where this migration has been implemented needs to be
        bigger than the version the user executed specter for the first time
        """
        return self.__class__.should_execute_cls(
            self.specter_migrator.mig.first_event["version"], self.__class__.version
        )

    @classmethod
    def should_execute_cls(cls, version_first_started, migration_version):
        """This is the default method and separated out for better testability"""
        # We need to compare the version we have been started first with the version this migration
        # is suppose to migrate.
        # If version_first_started > self.version we don't need to execute this migration and will return False
        try:
            compare_version = compare(version_first_started, migration_version)
        except SpecterError:
            # if in doubt because versions are unparsable, execute!
            return True
        if compare_version == 1:
            return True
        elif compare_version == -1:
            return False
        elif compare_version == 0:
            return False

    @property
    def description(self) -> str:
        """Should return a (multiline) description of the migration which will get log.info() at execution-time"""
        # return """A dummy migration:
        #     * "foo" changed from str to list[str] to support multiple foos (PR #1234)
        #     * "bar" renamed to "thing" (PR #1235)
        #     * "blah" removed; no longer needed because... (PR #1236)
        # """
        raise Exception(
            "Must write a description of the changes the migration implements."
        )

    def execute(self):
        logger.error(
            f"This migration seem to have forgotten orverriding the execute-method!"
        )


class SpecterMigrator:
    """A Class managing Migrations. Not calling it managers, as this is reserved for Instances attached to the specter instance"""

    def __init__(self, data_folder):
        version = VersionChecker(specter=self)
        self.current_binary_version = version._get_current_version()
        self.current_data_version = "unknown"
        self.data_folder = data_folder
        self.mig = MigDataManager(data_folder)
        if self.mig.latest_event["version"] != self.current_binary_version:
            logger.info(
                f"A new version has been started compared to last time: {self.current_binary_version}"
            )
            self.mig.create_new_event(self.current_binary_version)
        logger.debug(f"Initiated SpecterMigrator({self.mig})")

    def plan_migration(self):
        """Returns a list of instances from all the migration_1234-classes which hasn't been
        executed yet (according to migration_data.json)
        """
        migration_objects_list = []
        # The path where all the migrations are located:
        package_dir = str(Path(Path(__file__).resolve().parent, "migrations").resolve())
        for migration_class in self.get_migration_classes():
            migration_obj = migration_class(self)

            migration_id = SpecterMigrator.calculate_id(migration_obj)
            if not self.mig.has_migration_executed(migration_id):
                if migration_obj.should_execute():
                    logger.debug(
                        f"Adding class {migration_class.__name__} to list of planned migrations"
                    )
                    migration_objects_list.append(migration_obj)
                else:
                    logger.debug(
                        f"Skipping class {migration_class.__name__} because of 'should_execute' False"
                    )
        return migration_objects_list

    def execute_migrations(self, migration_object_list=None):
        if migration_object_list == None:
            migration_object_list = self.plan_migration()
        if not migration_object_list:
            logger.info("No Migrations to execute!")
        else:
            for object in migration_object_list:
                exec_id = SpecterMigrator.calculate_id(object)
                try:
                    logger.info(f"  --> Starting Migration {object.__class__.__name__}")
                    logger.info(object.description)
                    self.mig.create_new_exec_log(exec_id, self.current_binary_version)
                    object.execute()
                    logger.info(
                        f"  --> Completed Migration {object.__class__.__name__}"
                    )
                    self.mig.set_execution_log_status(exec_id, "completed")
                except Exception as e:
                    logger.error(
                        f"  --> Error in Migration {object.__class__.__name__}"
                    )
                    logger.exception(e)
                    self.mig.set_execution_log_status(exec_id, "error")
                    self.mig.set_Execution_log_error_msg(exec_id, str(e))

    @classmethod
    def calculate_id(cls, migration_obj):
        """Extract the id from ojects derived from classes like SpecterMigration_0123"""
        prefix, id = migration_obj.__class__.__name__.split("_")
        assert prefix == "SpecterMigration"
        return int(id)

    @classmethod
    def get_migration_classes(cls):
        """Returns all subclasses of class SpecterMigration"""
        class_list = []
        # The path where all the migrations are located:
        package_dir = str(Path(Path(__file__).resolve().parent, "migrations").resolve())
        for (_, module_name, _) in iter_modules(
            [package_dir]
        ):  # import the module and iterate through its attributes
            module = import_module(
                f"cryptoadvance.specter.util.migrations.{module_name}"
            )
            logger.info("Collecting possible migrations ...")
            for attribute_name in dir(module):
                attribute = getattr(module, attribute_name)
                if isclass(attribute):
                    if (
                        issubclass(attribute, SpecterMigration)
                        and not attribute.__name__ == "SpecterMigration"
                    ):
                        class_list.append(attribute)
        return class_list


class MigDataManager(GenericDataManager):
    """A class handling the data for the SpecterMigrator in migration_data.json with some convenience-methods"""

    @classmethod
    def initial_data(cls):
        return {
            "events": [],  # contains elements like {"timestamp": someTimestamp, "version": "v1.5.0"}
            "migration_executions": [],  # contains elements like {"timestamp": someTimestamp, "migration_id": 0}
        }

    name_of_json_file = "migration_data.json"

    def __init__(self, data_folder):
        # creating folders if they don't exist
        # Usually Specter is doing that but in the case of tests, we're instantiated before Specter
        if not os.path.isdir(data_folder):
            os.makedirs(data_folder)
        super().__init__(data_folder)
        logger.info(f"Initiated {self}")

    @property
    def events(self):
        """A list of events where one event represents the initial use of a new specter-version which is a different
        version than the one used before. Looks like: {"timestamp": someTimestamp, "version": "v1.5.0"}
        """
        return self.data["events"]

    @property
    def latest_event(self):
        if self.events:
            return self.events[-1]
        return {"timestamp": None, "version": None}

    @property
    def first_event(self):
        """Effectively the binary version the data has been created with FIRST"""
        if self.events:
            return self.events[0]
        raise SpecterError("You should not check the first_event before you've set it!")

    def create_new_event(self, version):
        timestamp = datetime.now()
        logger.debug(f"Creating new event with version {version} at {timestamp}")
        self.events.append({"timestamp": str(timestamp), "version": version})
        self._save()

    def create_new_exec_log(self, migration_id, executing_version):
        timestamp = datetime.now()
        logger.debug(
            f"Creating new Execution log with id {migration_id} at {timestamp}"
        )
        self.migration_executions.append(
            {
                "timestamp": str(timestamp),
                "migration_id": migration_id,
                "status": "started",
                "executing_version": executing_version,
            }
        )
        self._save()

    def set_execution_log_status(self, id, status):
        logger.debug(f"Setting execution log status of {id} to {status}")
        self._find_exec_log(id)["status"] = status
        self._save()

    def set_Execution_log_error_msg(self, id, msg):
        logger.debug(f"Setting execution log error_message of {id} to {msg}")
        self._find_exec_log(id)["error_msg"] = msg
        self._save()

    def _find_exec_log(self, id):
        for migration in self.migration_executions:
            if migration.get("migration_id") == id:
                return migration
        raise SpecterError(f"Can't find migration_execution with id {id}")

    @property
    def migration_executions(self):
        return self.data["migration_executions"]

    def has_migration_executed(self, migration_id):
        executed_list = [
            migration_execution.get("migration_id")
            for migration_execution in self.migration_executions
        ]
        logger.debug(f"Executed migration_classes ids: {executed_list}")
        return migration_id in executed_list

    def __repr__(self):
        return f"MigDataManager({self.data_file} events:{len(self.events)} execs:{len(self.migration_executions)} )"



import json
import logging
import os
from importlib import import_module
from inspect import isclass
from pathlib import Path
from pkgutil import iter_modules
import sys
from typing import Dict, List

from cryptoadvance.specter.config import ProductionConfig
from cryptoadvance.specter.managers.singleton import ConfigurableSingletonException
from cryptoadvance.specter.specter_error import SpecterError
from cryptoadvance.specter.user import User
from cryptoadvance.specter.util.reflection import get_template_static_folder
from flask import current_app as app
from flask import url_for
from flask.blueprints import Blueprint

from ..services.service import Service
from ..services import callbacks, ExtensionException
from ..services.service_encrypted_storage import (
    ServiceEncryptedStorageManager,
    ServiceUnencryptedStorageManager,
)
from ..util.reflection import (
    _get_module_from_class,
    get_classlist_of_type_clazz_from_modulelist,
    get_package_dir_for_subclasses_of,
    get_subclasses_for_clazz,
    get_subclasses_for_clazz_in_cwd,
)
from ..util.reflection_fs import search_dirs_in_path

logger = logging.getLogger(__name__)


class ServiceManager:
    """Loads support for all Services it auto-discovers."""

    def __init__(self, specter, devstatus_threshold):
        self.specter = specter
        specter.ext = {}
        self.devstatus_threshold = devstatus_threshold

        # Each Service class is stored here, keyed on its Service.id str
        self._services: Dict[str, Service] = {}
        logger.info("----> starting service discovery Static")
        # How do we discover services? Two configs are relevant:
        # * SERVICES_LOAD_FROM_CWD (boolean, CWD is current working directory)
        # * EXTENSION_LIST (array of Fully Qualified module strings like ["cryptoadvance.specter.services.swan.service"])
        # Ensuring security (especially for the CWD) is NOT done here but
        # in the corresponding (Production)Config
        logger.debug(f"EXTENSION_LIST = {app.config.get('EXTENSION_LIST')}")
        class_list = get_classlist_of_type_clazz_from_modulelist(
            Service, app.config.get("EXTENSION_LIST", [])
        )

        if app.config.get("SERVICES_LOAD_FROM_CWD", False):
            logger.info("----> starting service discovery dynamic")
            class_list.extend(get_subclasses_for_clazz_in_cwd(Service))
        else:
            logger.info("----> skipping service discovery dynamic")
        logger.info("----> starting service loading")
        class_list = set(class_list)  # remove duplicates (shouldn't happen but  ...)
        for clazz in class_list:
            compare_map = {"alpha": 1, "beta": 2, "prod": 3}
            if compare_map[self.devstatus_threshold] <= compare_map[clazz.devstatus]:
                # First configure the service
                self.configure_service_for_module(clazz)
                # Now activate it
                self._services[clazz.id] = clazz(
                    active=clazz.id in self.specter.config.get("services", []),
                    specter=self.specter,
                )
                self.specter.ext[clazz.id] = self._services[clazz.id]
                # maybe register the blueprint
                self.register_blueprint_for_ext(clazz, self._services[clazz.id])
                logger.info(f"Service {clazz.__name__} activated ({clazz.devstatus})")
            else:
                logger.info(
                    f"Service {clazz.__name__} not activated due to devstatus ( {self.devstatus_threshold} > {clazz.devstatus} )"
                )

        # Configure and instantiate the one and only ServiceEncryptedStorageManager
        try:
            ServiceEncryptedStorageManager.configure_instance(
                specter.data_folder, specter.user_manager
            )
        except ConfigurableSingletonException as e:
            # Test suite triggers multiple calls; ignore for now.
            pass

        specter.service_unencrypted_storage_manager = ServiceUnencryptedStorageManager(
            specter.user_manager, specter.data_folder
        )

        logger.info("----> finished service processing")
        self.execute_ext_callbacks("afterServiceManagerInit")

    @classmethod
    def register_blueprint_for_ext(cls, clazz, ext):
        if not clazz.has_blueprint:
            return
        if hasattr(clazz, "blueprint_modules"):
            controller_modules = clazz.blueprint_modules
            setattr(clazz, "blueprints", {})
        elif hasattr(clazz, "blueprint_module"):
            controller_modules = {"default": clazz.blueprint_module}
        else:
            import_name = f"cryptoadvance.specter.services.{clazz.id}.service"
            controller_modules = controller_modules = {
                "default": f"cryptoadvance.specter.services.{clazz.id}.controller"
            }

        only_one_blueprint = len(controller_modules.items()) == 1

        def inject_stuff():
            """Can be used in all jinja2 templates"""
            return dict(specter=app.specter, service=ext)

        if "default" not in controller_modules.keys():
            raise SpecterError(
                "You need at least one Blueprint, with the key 'default'. It will be used to link to your UI"
            )

        for bp_key, bp_value in controller_modules.items():
            if bp_key == "":
                raise SpecterError("Empty keys are not allowed in the blueprints map")
            middple_part = "" if bp_key == "default" else f"{bp_key}_"
            bp_name = f"{clazz.id}_{middple_part}endpoint"
            logger.debug(
                f"  Creating blueprint with name {bp_name} (middle_part:{middple_part}:"
            )
            bp = Blueprint(
                f"{clazz.id}_{middple_part}endpoint",
                bp_value,
                template_folder=get_template_static_folder("templates"),
                static_folder=get_template_static_folder("static"),
            )
            if only_one_blueprint:
                setattr(clazz, "blueprint", bp)
            else:
                clazz.blueprints[bp_key] = bp
            bp.context_processor(inject_stuff)

            # Import the controller for this service
            logger.info(f"  Loading Controller {bp_value}")

            try:
                controller_module = import_module(bp_value)
            except ModuleNotFoundError as e:
                raise Exception(
                    f"""
                    There was an issue finding a controller module:
                    {e}
                    That module was specified in the Service class of service {clazz.id}
                    check that specification in {clazz.__module__}
                """
                )

            # finally register the blueprint
            if clazz.isolated_client:
                ext_prefix = app.config["ISOLATED_CLIENT_EXT_URL_PREFIX"]
            else:
                ext_prefix = app.config["EXT_URL_PREFIX"]

            try:
                bp_postfix = "" if only_one_blueprint else f"/{bp_key}"
                if (
                    app.testing
                    and len(
                        [vf for vf in app.view_functions if vf.startswith(clazz.id)]
                    )
                    <= 1
                ):  # the swan-static one
                    # Yet again that nasty workaround which has been described in the archblog.
                    # The easy variant can be found in server.py
                    # The good news is, that we'll only do that for testing
                    import importlib

                    logger.info("Reloading Extension controller")
                    importlib.reload(controller_module)
                    app.register_blueprint(
                        bp, url_prefix=f"{ext_prefix}/{clazz.id}{bp_postfix}"
                    )
                else:
                    app.register_blueprint(
                        bp, url_prefix=f"{ext_prefix}/{clazz.id}{bp_postfix}"
                    )
                logger.info(f"  Mounted {bp} to {ext_prefix}/{clazz.id}{bp_postfix}")
            except AssertionError as e:
                if str(e).startswith("A name collision"):
                    raise SpecterError(
                        f"""
                    There is a name collision for the {clazz.blueprint.name}. \n
                    This is probably because you're running in DevelopementConfig and configured
                    the extension at the same time in the EXTENSION_LIST which currently loks like this:
                    {app.config['EXTENSION_LIST']})
                    """
                    )

    @classmethod
    def configure_service_for_module(cls, clazz):
        """searches for ConfigClasses in the module-Directory and merges its config in the global config"""
        try:
            module = import_module(f"cryptoadvance.specter.services.{clazz.id}.config")
        except ModuleNotFoundError:
            # maybe the other style:
            org = clazz.__module__.split(".")[0]
            try:
                module = import_module(f"{org}.specterext.{clazz.id}.config")
            except ModuleNotFoundError:
                logger.warning(
                    f"Service {clazz.id} does not have a service Configuration! Skipping!"
                )
                return
        main_config_clazz_name = app.config.get("SPECTER_CONFIGURATION_CLASS_FULLNAME")
        main_config_clazz_slug = main_config_clazz_name.split(".")[-1]
        potential_config_classes = []
        for attribute_name in dir(module):
            attribute = getattr(module, attribute_name)
            if isclass(attribute):
                clazz = attribute
                potential_config_classes.append(clazz)
                if clazz.__name__.endswith(
                    main_config_clazz_slug
                ):  # e.g. BaseConfig or DevelopmentConfig
                    cls.import_config(clazz)
                    return

        config_module = import_module(".".join(main_config_clazz_name.split(".")[0:-1]))

        config_clazz = getattr(config_module, main_config_clazz_slug)
        config_candidate_class = config_clazz.__bases__[0]
        while config_candidate_class != object:
            for clazz in potential_config_classes:
                if clazz.__name__.endswith(config_candidate_class.__name__):
                    cls.import_config(clazz)
                    return
            config_candidate_class = config_candidate_class.__bases__[0]
        logger.warning(
            f"Could not find a configuration for Service {module}. Skipping configuration."
        )

    @classmethod
    def import_config(cls, clazz):
        logger.info(f"  Loading Service-specific configuration from {clazz}")
        for key in dir(clazz):
            if key.isupper():
                if app.config.get(key):
                    raise Exception(
                        f"Config {clazz} tries to override existing key {key}"
                    )
                app.config[key] = getattr(clazz, key)
                logger.debug(f"setting {key} = {app.config[key]}")

    def execute_ext_callbacks(self, callback_id, *args, **kwargs):
        """will execute the callback function for each extension which has defined that method
        the callback_id needs to be passed and specify why the callback has been called.
        It needs to be one of the constants defined in cryptoadvance.specter.services.callbacks
        """
        if callback_id not in dir(callbacks):
            raise Exception(f"Non existing callback_id: {callback_id}")
        # No debug statement here possible as this is called for every request and would flood the logs
        # logger.debug(f"Executing callback {callback_id}")
        for ext in self.services.values():
            if hasattr(ext, f"callback_{callback_id}"):
                getattr(ext, f"callback_{callback_id}")(*args, **kwargs)
            elif hasattr(ext, "callback"):
                ext.callback(callback_id, *args, **kwargs)

    @property
    def services(self) -> Dict[str, Service]:
        return self._services or {}

    @property
    def services_sorted(self):
        service_names = sorted(
            self._services, key=lambda s: self._services[s].sort_priority
        )
        return [self._services[s] for s in service_names]

    def user_has_encrypted_storage(self, user: User) -> bool:
        """Looks for any data for any service in the User's ServiceEncryptedStorage.
        This check works even if the user doesn't have their plaintext_user_secret
        available."""
        encrypted_data = (
            ServiceEncryptedStorageManager.get_instance().get_raw_encrypted_data(user)
        )
        print(f"encrypted_data: {encrypted_data} for {user}")
        return encrypted_data != {}

    def set_active_services(self, service_names_active):
        logger.debug(f"Setting these services active: {service_names_active}")
        self.specter.update_services(service_names_active)
        for _, ext in self.services.items():
            logger.debug(
                f"Setting service '{ext.id}' active to {ext.id in service_names_active}"
            )
            ext.active = ext.id in service_names_active

    def get_service(self, plugin_id: str) -> Service:
        """get an extension-instance by ID. Raises an ExtensionException if it doesn't find it."""
        if plugin_id not in self._services:
            raise ExtensionException(f"No such plugin: '{plugin_id}'")
        return self._services[plugin_id]

    def remove_all_services_from_user(self, user: User):
        """
        Clears User.services and `user_secret`; wipes the User's
        ServiceEncryptedStorage.
        """
        # Don't show any Services on the sidebar for the admin user
        user.services.clear()

        # Reset as if we never had any encrypted storage
        user.delete_user_secret(autosave=False)
        user.save_info()

        if self.user_has_encrypted_storage(user=user):
            # Encrypted Service data is now orphaned since there is no
            # password. So wipe it from the disk.
            ServiceEncryptedStorageManager.get_instance().delete_all_service_data(user)

    @classmethod
    def get_service_x_dirs(cls, x):
        """returns a list of package-directories which each represents a specific service.
        This is used EXCLUSIVELY by the pyinstaller-hook packaging specter to add templates/static
        When this gets called, CWD is ./pyinstaller
        """

        arr = [
            Path(Path(_get_module_from_class(clazz).__file__).parent, x)
            for clazz in get_subclasses_for_clazz(Service)
        ]
        logger.debug(f"Initial arr:")
        for element in arr:
            logger.debug(element)
        # /home/kim/src/specter-desktop/.buildenv/lib/python3.8/site-packages/cryptoadvance/specter/services/bitcoinreserve/templates
        # /home/kim/src/specter-desktop/.buildenv/lib/python3.8/site-packages/cryptoadvance/specter/services/swan/templates

        arr = [path for path in arr if path.is_dir()]
        # Those pathes are absolute. Let's make them relative:
        arr = [Path(*path.parts[-6:]) for path in arr]

        virtuelenv_path = os.path.relpath(os.environ["VIRTUAL_ENV"], ".")

        if os.name == "nt":
            virtualenv_search_path = Path(virtuelenv_path, "Lib")
        else:
            # let's calcultate so that we get something like:
            # virtualenv_search_path = Path("..", ".buildenv", "lib", "python3.8")
            site_package = [path for path in sys.path if "site-packages" in path][0]
            site_package = Path(virtuelenv_path, *(Path(site_package).parts[-3:-1]))
            virtualenv_search_path = site_package

        # ... and as the classes are in the .buildenv (see build-unix.sh) let's add ..
        arr = [Path(virtualenv_search_path, path) for path in arr]

        # Non internal-repo extensions sitting in org/specterext/... need to be added, too
        src_org_specterext_exts = search_dirs_in_path(
            virtualenv_search_path, return_without_extid=False
        )
        src_org_specterext_exts = [Path(path, x) for path in src_org_specterext_exts]

        arr.extend(src_org_specterext_exts)

        return arr

    @classmethod
    def get_service_packages(cls):
        """returns a list of strings containing the service-classes (+ controller/config-classes)
        This is used for hiddenimports in pyinstaller
        """
        arr = get_subclasses_for_clazz(Service)
        arr.extend(
            get_classlist_of_type_clazz_from_modulelist(
                Service, ProductionConfig.EXTENSION_LIST
            )
        )
        arr = [clazz.__module__ for clazz in arr]
        # Controller-Packagages from the services are not imported via the service but via the baseclass
        # Therefore hiddenimport don't find them. We have to do it here.
        cont_arr = [
            ".".join(package.split(".")[:-1]) + ".controller" for package in arr
        ]
        for controller_package in cont_arr:
            try:
                import_module(controller_package)
                arr.append(controller_package)
            except ImportError:
                pass
            except AttributeError:
                # something like:
                # AttributeError: type object 'BitcoinReserveService' has no attribute 'blueprint'
                # shows that the package is existing
                arr.append(controller_package)
            except RuntimeError:
                # something like
                # RuntimeError: Working outside of application context.
                # shows that the package is existing
                arr.append(controller_package)
        config_arr = [".".join(package.split(".")[:-1]) + ".config" for package in arr]
        for config_package in config_arr:
            try:
                import_module(config_package)
                arr.append(config_package)
            except ModuleNotFoundError as e:
                pass
        return arr



from . import DeviceTypes
from .hwi_device import HWIDevice
from hwilib.devices.trezor import TrezorClient


class Trezor(HWIDevice):
    device_type = DeviceTypes.TREZOR
    name = "Trezor"
    icon = "trezor_icon.svg"

    supports_hwi_toggle_passphrase = True
    supports_hwi_multisig_display_address = True

    @classmethod
    def get_client(cls, *args, **kwargs):
        return TrezorClient(*args, **kwargs)



import logging
from importlib import import_module
from inspect import isclass
import os
from pathlib import Path
import pkgutil
from pkgutil import iter_modules
import sys
from typing import List
from .common import camelcase2snake_case
from ..specter_error import SpecterError
from .shell import grep

from .reflection_fs import detect_extension_style_in_cwd, search_dirs_in_path

logger = logging.getLogger(__name__)


def _get_module_from_class(clazz):
    return import_module(clazz.__module__)


def get_template_static_folder(foldername):
    """convenience-method to return static/template methods taking pyinstaller into account
    foldername can be anything but probably most reasonable template or static
    """
    if getattr(sys, "frozen", False):
        # Best understood with the snippet below this section:
        # https://pyinstaller.readthedocs.io/en/v3.3.1/runtime-information.html#using-sys-executable-and-sys-argv-0
        x_folder = os.path.join(sys._MEIPASS, foldername)
    else:
        x_folder = foldername
    return x_folder


def get_package_dir_for_subclasses_of(clazz):
    """There are two occasions where this makes sense: migrations and service-classes. Depending on the clazz
    this function is returning package_directories where subclasses from clazz are supposed to be located
    * subclasses of SpecterMigration are located in a subpackage called migrations
    * subclasses of Servuce are located in subpackages of package cryptoadvance.specter.services
    I have to admit that this is not a pure util class as it's containing business-logic.
    It's no longer used for Service.
    """
    if clazz.__name__ == "SpecterMigration":
        return str(
            Path(
                Path(_get_module_from_class(clazz).__file__).resolve().parent,
                "migrations",
            ).resolve()
        )
    elif clazz.__name__ == "Service":
        return str(
            Path(
                import_module("cryptoadvance.specter.services").__file__
            ).parent.resolve()
        )
    # This is mainly for testing purposes for now
    elif clazz.__name__ == "Device":
        return str(
            Path(
                import_module("cryptoadvance.specter.devices").__file__
            ).parent.resolve()
        )
    raise SpecterError("Unknown Class: {clazz}")


# --------------- static discovery ------------------------------


def get_classlist_of_type_clazz_from_modulelist(clazz, modulelist):
    """A helper method converting a List of modules as described in config.py
    into a List of classes. In order to make that more util-like, you
    have to pass the the class you're searching for in the modules
    """
    class_list = []
    for fq_module_name in modulelist:
        try:
            module = import_module(fq_module_name)
        except ModuleNotFoundError as e:
            # ToDo: make it somehow clear where specific extensions are coming from: external or within same repo
            logger.warn(
                f"""
                Module {fq_module_name}  could not be found. This could have these reasons:
                * You might have forgot to: 
                    pip3 install yourPackage
                * You're trying to start the ProductionConfig in a Development Environment. 
                    If you checked out the specter-Sourcecode, you should start specter like this:
                    python3 -m cryptoadvance.specter server --config DevelopmentConfig --debug"""
            )
            continue
        logger.debug(f"Imported {fq_module_name}")
        for attribute_name in dir(module):
            attribute = getattr(module, attribute_name)
            if isclass(attribute):
                if (
                    issubclass(attribute, clazz)
                    and not attribute.__name__ == clazz.__name__
                ):
                    logger.debug(f"Adding {attribute} to {class_list}")
                    class_list.append(attribute)
                    logger.info(f"  Found class {attribute.__name__}")
    return class_list


def get_subclasses_for_clazz_in_cwd(clazz, cwd=".") -> List[type]:
    """Returns all subclasses of class clazz located in the CWD if the cwd
    is not a specter-desktop dev-env-kind-of-dir or contains any .py-file
    So
    """
    package_dirs = []
    # security first! No dynamic loading in app-images
    if getattr(sys, "frozen", False):
        return []

    # if not testing but in a folder which looks like specter-desktop/src --> No dynamic extensions
    if "PYTEST_CURRENT_TEST" not in os.environ:
        # Hmm, need a better way to detect a specter-desktop-sourcedir
        try:
            if grep("./setup.py", 'name="cryptoadvance.specter",'):
                return []
        except FileNotFoundError:
            pass

    # Depending on the style we either add "." or "./src" to the searchpath

    extension_style = detect_extension_style_in_cwd()
    if extension_style == "adhoc":
        package_dirs.append(Path("."))
    elif extension_style == "publish-ready":
        package_dirs.extend(search_dirs_in_path(Path("./src")))
    elif extension_style == "specter-desktop":
        if "PYTEST_CURRENT_TEST" in os.environ:
            # I admit, ugly hack
            logger.info("We're in testing mode. Adding CWD to searchpath")
            package_dirs.append(Path("./src"))
        else:
            raise Exception(f"This should not happen")
    logger.info(f"Detected Extension-style: {extension_style}")
    logger.info(f"We'll search in those package_dirs {package_dirs}")
    return get_subclasses_for_clazz(clazz, package_dirs)


def get_subclasses_for_clazz(clazz, package_dirs: List[str] = None):
    """Returns all subclasses of class clazz located in the CWD
    potentially add additional_packagedirs which is usefull for
    calculating pyinstaller hiddenimports
    """
    if package_dirs == None:
        package_dirs = [get_package_dir_for_subclasses_of(clazz)]
    class_list = []
    logger.info(
        f"Collecting subclasses of {clazz.__name__} in {' '.join([ str(dir) for dir in package_dirs]) }..."
    )
    for (importer, module_name, is_pkg) in iter_modules(
        [str(dir) for dir in package_dirs]
    ):  # import the module and iterate through its attributes
        # skip known redherrings
        if module_name in ["callbacks"]:
            continue
        logger.debug(
            f"Iterating on importer={importer} , module_name={module_name} is_pkg={is_pkg}"
        )
        if clazz.__name__ == "Service":
            try:

                module = import_module(
                    f"cryptoadvance.specter.services.{module_name}.service"
                )
                logger.debug(
                    f"  Imported cryptoadvance.specter.services.{module_name}.service"
                )
            except ModuleNotFoundError:
                # Ignore the stuff lying around in cryptoadvance/specter/services
                if importer.path.endswith(
                    os.path.sep.join(["cryptoadvance", "specter", "services"])
                ):
                    continue
                try:
                    module = import_module(f"{module_name}.service")
                    logger.debug(f"  Imported {module_name}.service")
                except ModuleNotFoundError as e:
                    try:
                        # Another style is orgname.specterext.extensionid, for that we have to guess the orgname:
                        orgname = str(importer).split(os.path.sep)[-2]
                        logger.debug(f"guessing orgname: {orgname}")
                        module = import_module(
                            f"{orgname}.specterext.{module_name}.service"
                        )
                        logger.debug(
                            f"  Imported {orgname}.specterext.{module_name}.service"
                        )
                    except ModuleNotFoundError as e:
                        if module_name in str(e.name) or orgname in str(e.name):
                            raise Exception(
                                f"""
                        While iterating over {importer} for module {module_name}, 
                        a Service implementation could not be found in this places:
                        * cryptoadvance.specter.services.{module_name}.service
                        * {module_name}.service
                        * {orgname}.specterext.{module_name}.service
                        Maybe you did forget to do this:
                        $ pip3 install -e .

                        OR The Module has been found, but had issues finding Modules itself

                        {e}
                        """
                            )
                        else:
                            raise e
        elif clazz.__name__ == "SpecterMigration":
            module = import_module(
                f"cryptoadvance.specter.util.migrations.{module_name}"
            )
        else:
            try:
                module = import_module(
                    f"{module_name}.{camelcase2snake_case(clazz.__name__)}"
                )
                logger.debug(
                    f"Imported {module_name}.{camelcase2snake_case(clazz.__name__)}"
                )
            except ModuleNotFoundError as e:
                logger.debug(
                    f"No Service Impl found in {module_name}.service. Skipping!"
                )
                continue
        for attribute_name in dir(module):
            attribute = getattr(module, attribute_name)
            if isclass(attribute):
                if (
                    issubclass(attribute, clazz)
                    and not attribute.__name__ == clazz.__name__
                ):
                    class_list.append(attribute)
                    logger.info(f"  Found class {attribute.__name__}")
    return class_list



import copy
import logging
from io import BytesIO

from embit.descriptor.checksum import add_checksum
from embit.ec import PrivateKey, PublicKey
from embit.hashes import tagged_hash
from embit.liquid import finalizer, slip77
from embit.liquid.addresses import addr_decode
from embit.liquid.addresses import address as liquid_address
from embit.liquid.descriptor import LDescriptor
from embit.liquid.networks import get_network
from embit.liquid.pset import PSET, PSBTError
from embit.liquid.transaction import LTransaction, unblind
from embit.psbt import read_string

from ..rpc import BitcoinRPC, RpcError
from .util.pset import to_canonical_pset

logger = logging.getLogger(__name__)


class LiquidRPC(BitcoinRPC):
    """
    This class introduces fixes to Elements RPC that are broken in current Elements Core.
    It also adds custom fields that are normaly not present in the final blind pset, but
    that we find useful.

    In particular this class:
    - combines walletcreatefundedpsbt and blindpsbt calls into one call by default (walletcreatefundedpsbt)
    - TODO: adds support for assets in walletcreatefundedpsbt
    - TODO: adds custom fields to pset:
      - blinding pubkeys for HW recv addr verification
      - ecdh nonce for deterministic range proof
      - extra data that should be encoded in the change rangeproof
    """

    _master_blinding_key = None
    _asset_labels = None

    @property
    def asset_labels(self):
        if self._asset_labels is None:
            self._asset_labels = super().__getattr__("dumpassetlabels")()
        return self._asset_labels

    @property
    def master_blinding_key(self):
        if self._master_blinding_key is None:
            self._master_blinding_key = PrivateKey(
                bytes.fromhex(self.dumpmasterblindingkey())
            )
        return self._master_blinding_key

    def dumpassetlabels(self, *args, **kwargs):
        # inject cache
        res = super().__getattr__("dumpassetlabels")()
        self._asset_labels = res
        return res

    def importmasterblindingkey(self, *args, **kwargs):
        # clear cache
        self._master_blinding_key = None
        return super().__getattr__("importmasterblindingkey")(*args, **kwargs)

    def _patch_assetlabels(self, obj):
        # check if assets are labeled and replace with real assetsids
        if isinstance(obj, dict):
            for asset in list(obj.keys()):
                if asset in self.asset_labels and asset != "bitcoin":
                    assetid = self.asset_labels[asset]
                    obj[assetid] = obj[asset]
                    del obj[asset]

    def getbalance(
        self,
        dummy="*",
        minconf=0,
        include_watchonly=True,
        avoid_reuse=False,
        assetlabel="bitcoin",  # pass None to get all
        **kwargs,
    ):
        """
        Bitcoin-like getbalance rpc call without assets,
        set assetlabel=None to get balances for all assets
        """
        args = [dummy, minconf, include_watchonly, avoid_reuse]
        # if assetlabel is explicitly set to None - return all assets
        if assetlabel is not None:
            args.append(assetlabel)
        res = super().__getattr__("getbalance")(*args, **kwargs)
        self._patch_assetlabels(res)
        return res

    def _patch_descriptors(self, arr):
        # used by importmulti and importdescriptors
        descs = [LDescriptor.from_string(a["desc"].split("#")[0]) for a in arr]
        blinded = [d.is_blinded for d in descs]
        bkey = None
        if any(blinded) and not all(blinded):
            raise RpcError("All descriptors should be either blinded or not")
        if all(blinded):
            bkeys = {d.blinding_key.key.secret for d in descs}
            if len(bkeys) > 1:
                raise RpcError("All descriptors should use the same blinding key")
            bkey = bkeys.pop().hex()
            arr = [a.copy() for a in arr]
            for i, a in enumerate(arr):
                descs[i].blinding_key = None
                a["desc"] = add_checksum(str(descs[i]))
        return arr, bkey

    def importdescriptors(self, arr, *args, **kwargs):
        arr, bkey = self._patch_descriptors(arr)
        res = super().__getattr__("importdescriptors")(arr, *args, **kwargs)
        if bkey is not None:
            self.importmasterblindingkey(bkey, **kwargs)
        return res

    def importmulti(self, arr, *args, **kwargs):
        arr, bkey = self._patch_descriptors(arr)
        res = super().__getattr__("importmulti")(arr, *args, **kwargs)
        if bkey is not None:
            self.importmasterblindingkey(bkey, **kwargs)
        return res

    def getbalances(self, assetlabel="bitcoin", **kwargs):
        """
        Bitcoin-like getbalance rpc call without assets,
        set assetlabel=None to get balances for all assets
        """
        res = super().__getattr__("getbalances")(**kwargs)
        # if assetlabel is explicitly set to None - return as is
        if assetlabel is None:
            for k in res:
                for kk in res[k]:
                    self._patch_assetlabels(res[k][kk])
            return res
        # otherwise get balances for a particular assetlabel
        for k in res:
            for kk in res[k]:
                v = res[k][kk].get(assetlabel, 0)
                res[k][kk] = v
        return res

    def getreceivedbyaddress(self, address, minconf=1, assetlabel="bitcoin", **kwargs):
        args = [address, minconf]
        if assetlabel is not None:
            args.append(assetlabel)
        return super().__getattr__("getreceivedbyaddress")(*args, **kwargs)

    def walletcreatefundedpsbt(
        self, inputs, outputs, locktime=0, options={}, *args, blind=True, **kwargs
    ):
        """
        Creates and blinds an Elements PSBT transaction.
        Arguments:
        1. inputs: [{txid, vout[, sequence, pegin stuff]}]
        2. outputs: [{address: amount, "asset": asset}, ...] # TODO: add assets support
        3. locktime = 0
        4. options {includeWatching, changeAddress, subtractFeeFromOutputs,
                    replaceable, add_inputs, feeRate, fee_rate, changeAddresses}
        5. bip32 derivations
        6. solving data
        7. blind = True - Specter-LiquidRPC specific thing - blind transaction after creation
        """
        options = copy.deepcopy(options)
        change_addresses = (
            options.pop("changeAddresses") if "changeAddresses" in options else None
        )
        destinations = []
        for o in outputs:
            for k in o:
                if k != "asset":
                    destinations.append(addr_decode(k)[0])
        res = super().__getattr__("walletcreatefundedpsbt")(
            inputs, outputs, locktime, options, *args, **kwargs
        )
        psbt = res.get("psbt", None)

        # remove zero-output (bug in Elements)
        # TODO: remove after release
        if psbt:
            try:
                tx = PSET.from_string(psbt)
                # check if there are zero outputs
                has_zero = len([out for out in tx.outputs if out.value == 0]) > 0
                has_blinded = any([out.blinding_pubkey for out in tx.outputs])
                logger.error(has_zer, has_blinded)
                if has_blinded and has_zero:
                    tx.outputs = [out for out in tx.outputs if out.value > 0]
                    psbt = str(tx)
                    res["psbt"] = psbt
            except:
                pass

        # replace change addresses from the transactions if we can
        if change_addresses and psbt:
            try:
                tx = PSET.from_string(psbt)
                cur = 0
                for out in tx.outputs:
                    # fee
                    if out.script_pubkey.data == b"":
                        continue
                    # not change for sure
                    if not out.bip32_derivations:
                        continue
                    # do change replacement
                    if out.script_pubkey not in destinations:
                        sc, bkey = addr_decode(change_addresses[cur])
                        cur += 1
                        out.script_pubkey = sc
                        out.blinding_pubkey = bkey.sec() if bkey else None
                        out.bip32_derivations = {}
                        out.redeem_script = None
                        out.witness_script = None
                # fill derivation info
                patched = (
                    super().__getattr__("walletprocesspsbt")(str(tx), False).get("psbt")
                )
                patchedtx = PSET.from_string(patched)
                assert len(tx.outputs) == len(patchedtx.outputs)
                for out1, out2 in zip(tx.outputs, patchedtx.outputs):
                    # fee
                    if out1.script_pubkey.data == b"":
                        continue
                    # not change for sure
                    if not out2.bip32_derivations:
                        continue
                    # do change replacement
                    if out1.script_pubkey not in destinations:
                        out1.bip32_derivations = out2.bip32_derivations
                        out1.redeem_script = out2.redeem_script
                        out1.witness_script = out2.witness_script

                res["psbt"] = str(tx)
            except Exception as e:
                logger.error(e)
                raise e

        psbt = res.get("psbt", None)
        # check if we should blind the transaction
        if psbt and blind:
            # check that change is also blinded - fixes a bug in pset branch
            tx = PSET.from_string(psbt)
            changepos = res.get("changepos", None)
            # no change output
            if changepos < 0:
                changepos = None

            # generate all blinding stuff ourselves in deterministic way
            tx.unblind(
                self.master_blinding_key
            )  # get values and blinding factors for inputs
            seed = tagged_hash("liquid/blinding_seed", self.master_blinding_key.secret)
            try:
                tx.blind(seed)  # generate all blinding factors etc
                # proprietary fields for Specter - 00 is global blinding seed
                tx.unknown[b"\xfc\x07specter\x00"] = seed
            except PSBTError:
                seed = None

            # reblind and encode nonces in change output
            if seed and changepos is not None:
                txseed = tx.txseed(seed)
                # blinding seed to calculate per-output nonces
                message = b"\x01\x00\x20" + txseed
                for i, out in enumerate(tx.outputs):
                    # skip unblinded and change address itself
                    if out.blinding_pubkey is None or i == changepos:
                        continue
                    # key 01<i> is blinding pubkey for output i
                    message += b"\x05\x01" + i.to_bytes(4, "little")
                    # message is blinding pubkey
                    message += bytes([len(out.blinding_pubkey)]) + out.blinding_pubkey
                # extra message for rangeproof - proprietary field
                tx.outputs[changepos].unknown[b"\xfc\x07specter\x01"] = message
                # re-generate rangeproof with extra message
                nonce = tagged_hash(
                    "liquid/range_proof", txseed + changepos.to_bytes(4, "little")
                )
                tx.outputs[changepos].reblind(nonce, extra_message=message)

            res["psbt"] = str(tx)
        return res

    def walletprocesspsbt(self, psbt, *args, **kwargs):
        try:
            if self.getwalletinfo().get("private_keys_enabled", False):
                psbt = to_canonical_pset(psbt)
        except Exception as e:
            logger.warn(f"Failed to clean psbt: {e}")
        return super().__getattr__("walletprocesspsbt")(psbt, *args, **kwargs)

    def finalizepsbt(self, psbt, *args, **kwargs):
        psbt = to_canonical_pset(psbt)
        res = super().__getattr__("finalizepsbt")(psbt, *args, **kwargs)
        if res["complete"] == False:
            try:
                # try using our finalizer
                tx = finalizer.finalize_psbt(PSET.from_string(psbt))
                if tx and self.testmempoolaccept([str(tx)]):
                    return {"complete": True, "hex": str(tx)}
            except Exception as e:
                logger.exception(e)
        return res

    def combinepsbt(self, psbts, *args, **kwargs):
        if len(psbts) == 0:
            raise RpcError("Provide at least one psbt")
        tx = PSET.from_string(psbts[0])
        for b64 in psbts[1:]:
            t2 = PSET.from_string(b64)
            tx.version = tx.version or t2.version
            tx.tx_version = tx.tx_version or t2.tx_version
            tx.locktime = tx.locktime or t2.locktime
            tx.xpubs.update(t2.xpubs)
            tx.unknown.update(t2.unknown)

            for inp1, inp2 in zip(tx.inputs, t2.inputs):
                inp1.value = inp1.value or inp2.value
                inp1.value_blinding_factor = (
                    inp1.value_blinding_factor or inp2.value_blinding_factor
                )
                inp1.asset = inp1.asset or inp2.asset
                inp1.asset_blinding_factor = (
                    inp1.asset_blinding_factor or inp2.asset_blinding_factor
                )
                inp1.txid = inp1.txid or inp2.txid
                inp1.vout = inp1.vout or inp2.vout
                inp1.sequence = inp1.sequence or inp2.sequence
                inp1.non_witness_utxo = inp1.non_witness_utxo or inp2.non_witness_utxo
                inp1.sighash_type = inp1.sighash_type or inp2.sighash_type
                inp1.redeem_script = inp1.redeem_script or inp2.redeem_script
                inp1.witness_script = inp1.witness_script or inp2.witness_script
                inp1.final_scriptsig = inp1.final_scriptsig or inp2.final_scriptsig
                inp1.final_scriptwitness = (
                    inp1.final_scriptwitness or inp2.final_scriptwitness
                )
                inp1.partial_sigs.update(inp2.partial_sigs)
                inp1.bip32_derivations.update(inp2.bip32_derivations)
                inp1.unknown.update(inp2.unknown)
                inp1.range_proof = inp1.range_proof or inp2.range_proof

            for out1, out2 in zip(tx.outputs, t2.outputs):
                out1.value_commitment = out1.value_commitment or out2.value_commitment
                out1.value_blinding_factor = (
                    out1.value_blinding_factor or out2.value_blinding_factor
                )
                out1.asset_commitment = out1.asset_commitment or out2.asset_commitment
                out1.asset_blinding_factor = (
                    out1.asset_blinding_factor or out2.asset_blinding_factor
                )
                out1.range_proof = out1.range_proof or out2.range_proof
                out1.surjection_proof = out1.surjection_proof or out2.surjection_proof
                out1.ecdh_pubkey = out1.ecdh_pubkey or out2.ecdh_pubkey
                out1.blinding_pubkey = out1.blinding_pubkey or out2.blinding_pubkey
                out1.asset = out1.asset or out2.asset

                out1.value = out1.value or out2.value
                out1.script_pubkey = out1.script_pubkey or out2.script_pubkey
                out1.unknown = out1.unknown or out2.unknown
                out1.redeem_script = out1.redeem_script or out2.redeem_script
                out1.witness_script = out1.witness_script or out2.witness_script
                out1.bip32_derivations.update(out2.bip32_derivations)
                out1.unknown.update(out2.unknown)
        return str(tx)

    def decoderawtransaction(self, tx):
        blinded = super().__getattr__("decoderawtransaction")(tx)
        try:
            unblinded = self.unblindrawtransaction(tx)["hex"]
            obj = super().__getattr__("decoderawtransaction")(unblinded)
            if "vsize" in blinded:
                obj["vsize"] = blinded["vsize"]
            if "size" in blinded:
                obj["size"] = blinded["size"]
            if "weight" in blinded:
                obj["weight"] = blinded["weight"]
            if "hex" in blinded:
                obj["hex"] = blinded["hex"]
        except Exception as e:
            logger.error(e)
            obj = blinded
        try:
            # unblind the rest of outputs
            b = LTransaction.from_string(tx)

            mbpk = self.master_blinding_key
            net = get_network(self.getblockchaininfo().get("chain"))

            outputs = obj["vout"]
            datas = []
            fee = 0
            # search for datas encoded in rangeproofs
            for i, out in enumerate(b.vout):
                o = outputs[i]
                if isinstance(out.value, int):
                    if "value" in o:
                        assert o["value"] == round(out.value * 1e-8, 8)
                    else:
                        o["value"] = round(out.value * 1e-8, 8)
                    if "asset" in o:
                        assert o["asset"] == bytes(reversed(out.asset[-32:])).hex()
                    else:
                        o["asset"] = bytes(reversed(out.asset[-32:])).hex()
                    try:
                        o["scriptPubKey"]["addresses"] = [
                            liquid_address(out.script_pubkey, network=net)
                        ]
                    except:
                        pass
                    if out.script_pubkey.data == b"":
                        # fee negative?
                        fee -= out.value

                pk = slip77.blinding_key(mbpk, out.script_pubkey)
                try:
                    res = out.unblind(pk.secret, message_length=1000)
                    value, asset, vbf, abf, extra, min_value, max_value = res
                    if "value" in o:
                        assert o["value"] == round(value * 1e-8, 8)
                    else:
                        o["value"] = round(value * 1e-8, 8)
                    if "asset" in o:
                        assert o["asset"] == bytes(reversed(asset[-32:])).hex()
                    else:
                        o["asset"] = bytes(reversed(asset[-32:])).hex()
                    try:
                        o["scriptPubKey"]["addresses"] = [
                            liquid_address(out.script_pubkey, pk, network=net)
                        ]
                    except:
                        pass
                    if len(extra.rstrip(b"\x00")) > 0:
                        datas.append(extra)
                except Exception as e:
                    pass

            # should be changed with seed from tx
            tx = PSET(b)
            seed = tagged_hash("liquid/blinding_seed", mbpk.secret)
            txseed = tx.txseed(seed)
            pubkeys = {}

            for extra in datas:
                s = BytesIO(extra)
                while True:
                    k = read_string(s)
                    if len(k) == 0:
                        break
                    v = read_string(s)
                    if k[0] == 1 and len(k) == 5:
                        idx = int.from_bytes(k[1:], "little")
                        pubkeys[idx] = v
                    elif k == b"\x01\x00":
                        txseed = v

            for i, out in enumerate(outputs):
                o = out
                if i in pubkeys and len(pubkeys[i]) in [33, 65]:
                    nonce = tagged_hash(
                        "liquid/range_proof", txseed + i.to_bytes(4, "little")
                    )
                    if b.vout[i].ecdh_pubkey == PrivateKey(nonce).sec():
                        try:
                            res = unblind(
                                pubkeys[i],
                                nonce,
                                b.vout[i].witness.range_proof.data,
                                b.vout[i].value,
                                b.vout[i].asset,
                                b.vout[i].script_pubkey,
                            )
                            value, asset, vbf, abf, extra, min_value, max_value = res
                            if "value" in o:
                                assert o["value"] == round(value * 1e-8, 8)
                            else:
                                o["value"] = round(value * 1e-8, 8)
                            if "asset" in o:
                                assert o["asset"] == bytes(reversed(asset[-32:])).hex()
                            else:
                                o["asset"] = bytes(reversed(asset[-32:])).hex()
                            try:
                                o["scriptPubKey"]["addresses"] = [
                                    liquid_address(
                                        b.vout[i].script_pubkey,
                                        PublicKey.parse(pubkeys[i]),
                                        network=net,
                                    )
                                ]
                            except:
                                pass
                        except Exception as e:
                            logger.warn(f"Failed at unblinding output {i}: {e}")
                    else:
                        logger.warn(f"Failed at unblinding: {e}")
            if fee != 0:
                obj["fee"] = round(-fee * 1e-8, 8)
        except Exception as e:
            logger.warn(f"Failed at unblinding transaction: {e}")
        return obj

    def __repr__(self) -> str:
        return f"<LiquidRpc {self.url}>"

    @classmethod
    def from_bitcoin_rpc(cls, rpc):
        """Convert BitcoinRPC to LiquidRPC"""
        return cls(
            user=rpc.user,
            password=rpc.password,
            host=rpc.host,
            port=rpc.port,
            protocol=rpc.protocol,
            path=rpc.path,
            timeout=cls.default_timeout,  # Elements is slower
            session=rpc.session,
            proxy_url=rpc.proxy_url,
            only_tor=rpc.only_tor,
        )



import binascii
import collections
import copy
import hashlib
import hmac
import json
import logging
import os
import six
import subprocess
import sys
from collections import OrderedDict
from mnemonic import Mnemonic
from embit.script import Script
from embit.psbt import PSBT
from embit.transaction import Transaction
from embit.liquid.pset import PSET
from embit.liquid.transaction import LTransaction
from embit.liquid.networks import NETWORKS
from .persistence import read_json_file, write_json_file
from .util.bcur import bcur_decode
import threading
from io import BytesIO
import re

logger = logging.getLogger(__name__)

# default lock for @locked()
defaultlock = threading.Lock()

EMOJIS = "😀😃😄😁😆😅😂🤣😊😇🙂🙃😉😌😍😘😗😙😚😋😜😝😛🤑🤗🤓😎🤡🤠😏😒😞😔😟😕🙁😣😖😫😩😤😠😡😶😐😑😯😦😧😮😲😵😳😱😨😰😢😥🤤😭😓😪😴🙄🤔🤥😬🤐🤢🤧😷🤒🤕😈👿👹👺💩👻💀☠️👽👾🤖🎃😺😸😹😻😼😽🙀😿😾👐🙌👏🙏🤝👍👎👊✊🤛🤜🤞✌️🤘👌👈👉👆👇☝️✋🤚🖐🖖👋🤙💪🖕✍️🤳💅🖖💄💋👄👅👂👃👣👁👀🗣👤👥👶👦👧👨👩👱👴👵👲👳👮👷💂🕵️👩‍⚕️👨‍⚕️👩‍🌾👨‍🌾👩‍🍳👨‍🍳👩‍🎓👨‍🎓👩‍🎤👨‍🎤👩‍🏫👨‍🏫👩‍🏭👨‍🏭👩‍💻👨‍💻👩‍💼👨‍💼👩‍🔧👨‍🔧👩‍🔬👨‍🔬👩‍🎨👨‍🎨👩‍🚒👨‍🚒👩‍✈️👨‍✈️👩‍🚀👨‍🚀👩‍⚖️👨‍⚖️🤶🎅👸🤴👰🤵👼🤰🙇💁🙅🙆🙋🙎🙍💇💆🕴💃🕺👯🚶🏃👫👭👬💑👩‍❤️‍👩👨‍❤️‍👨💏👩‍❤️‍💋‍👩👨‍❤️‍💋‍👨👪👨‍👩‍👧👨‍👩‍👧‍👦👨‍👩‍👦‍👦👨‍👩‍👧‍👧👩‍👩‍👦👩‍👩‍👧👩‍👩‍👧‍👦👩‍👩‍👦‍👦👩‍👩‍👧‍👧👨‍👨‍👦👨‍👨‍👧👨‍👨‍👧‍👦👨‍👨‍👦‍👦👨‍👨‍👧‍👧👩‍👦👩‍👧👩‍👧‍👦👩‍👦‍👦👩‍👧‍👧👨‍👦👨‍👧👨‍👧‍👦👨‍👦‍👦👨‍👧‍👧👚👕👖👔👗👙👘👠👡👢👞👟👒🎩🎓👑⛑🎒👝👛👜💼👓🕶🌂☂️🐶🐱🐭🐹🐰🦊🐻🐼🐨🐯🦁🐮🐷🐽🐸🐵🙊🙉🙊🐒🐔🐧🐦🐤🐣🐥🦆🦅🦉🦇🐺🐗🐴🦄🐝🐛🦋🐌🐚🐞🐜🕷🕸🐢🐍🦎🦂🦀🦑🐙🦐🐠🐟🐡🐬🦈🐳🐋🐊🐆🐅🐃🐂🐄🦌🐪🐫🐘🦏🦍🐎🐖🐐🐏🐑🐕🐩🐈🐓🦃🕊🐇🐁🐀🐿🐾🐉🐲🌵🎄🌲🌳🌴🌱🌿☘️🍀🎍🎋🍃🍂🍁🍄🌾💐🌷🌹🥀🌻🌼🌸🌺🌎🌍🌏🌕🌖🌗🌘🌑🌒🌓🌔🌚🌝🌞🌛🌜🌙💫⭐️🌟✨⚡️🔥💥☄️☀️🌤⛅️🌥🌦🌈☁️🌧⛈🌩🌨☃️⛄️❄️🌬💨🌪🌫🌊💧💦☔️🍏🍎🍐🍊🍋🍌🍉🍇🍓🍈🍒🍑🍍🥝🥑🍅🍆🥒🥕🌽🌶🥔🍠🌰🥜🍯🥐🍞🥖🧀🥚🍳🥓🥞🍤🍗🍖🍕🌭🍔🍟🥙🌮🌯🥗🥘🍝🍜🍲🍥🍣🍱🍛🍚🍙🍘🍢🍡🍧🍨🍦🍰🎂🍮🍭🍬🍫🍿🍩🍪🥛🍼☕️🍵🍶🍺🍻🥂🍷🥃🍸🍹🍾🥄🍴🍽⚽️🏀🏈⚾️🎾🏐🏉🎱🏓🏸🥅🏒🏑🏏⛳️🏹🎣🥊🥋⛸🎿⛷🏂🏋️🤺⛹️🏌️🏄🏊🚣🏇🚴🚵🎽🏅🎖🥇🥈🥉🏆🏵🎗🎫🎟🎪🎭🎨🎬🎤🎧🎼🎹🥁🎷🎺🎸🎻🎲🎯🎳🎮🎰🚗🚕🚙🚌🚎🏎🚓🚑🚒🚐🚚🚛🚜🛴🚲🛵🏍🚨🚔🚍🚘🚖🚡🚠🚟🚃🚋🚞🚝🚄🚅🚈🚂🚆🚇🚊🚉🚁🛩✈️🛫🛬🚀🛰💺🛶⛵️🛥🚤🛳⛴🚢⚓️🚧⛽️🚏🚦🚥🗺🗿🗽⛲️🗼🏰🏯🏟🎡🎢🎠⛱🏖🏝⛰🏔🗻🌋🏜🏕⛺️🛤🛣🏗🏭🏠🏡🏘🏚🏢🏬🏣🏤🏥🏦🏨🏪🏫🏩💒🏛⛪️🕌🕍🕋⛩🗾🎑🏞🌅🌄🌠🎇🎆🌇🌆🏙🌃🌌🌉🌁⌚️📱📲💻⌨️🖥🖨🖱🖲🕹🗜💽💾💿📀📼📷📸📹🎥📽🎞📞☎️📟📠📺📻🎙🎚🎛⏱⏲⏰🕰⌛️⏳📡🔋🔌💡🔦🕯🗑🛢💸💵💴💶💷💰💳💎⚖️🔧🔨⚒🛠⛏🔩⚙️⛓🔫💣🔪🗡⚔️🛡🚬⚰️⚱️🏺🔮📿💈⚗️🔭🔬🕳💊💉🌡🚽🚰🚿🛁🛀🛎🔑🗝🚪🛋🛏🛌🖼🛍🛒🎁🎈🎏🎀🎊🎉🎎🏮🎐✉️📩📨📧💌📥📤📦🏷📪📫📬📭📮📯📜📃📄📑📊📈📉🗒🗓📆📅📇🗃🗳🗄📋📁📂🗂🗞📰📓📔📒📕📗📘📙📚📖🔖🔗📎🖇📐📏📌📍📌🎌🏳️🏴🏁🏳️‍🌈✂️🖊🖋✒️🖌🖍📝✏️🔍🔎🔏🔐🔒🔓❤️💛💚💙💜🖤💔❣️💕💞💓💗💖💘💝🔈🔇🔉🔊🔔🔕📣📢👁‍🗨💬"


def get_asset_label(asset, known_assets={}):
    # TODO: lookup in the registry
    if asset in [None, "00" * 32, "ff" * 32]:
        return "???"
    if asset == "bitcoin":
        return "LBTC"
    if asset in known_assets:
        return known_assets[asset]
    icon = ""
    try:
        n = int.from_bytes(bytes.fromhex(asset), "little") % len(EMOJIS)
        icon = EMOJIS[n : n + 1]
    except:
        pass
    if len(asset) > 16:
        return f"{icon}"
    return asset


def add_dicts(target, item):
    for k, v in item.items():
        if isinstance(v, dict):
            target[k] = add_dicts(target.get(k, {}), v)
        else:
            # this will add ints, floats, strings etc
            target[k] = target.get(k, type(v)()) + v
    return target


def is_testnet(chain):
    return chain not in ["main", "liquidv1", "None", "none", None, ""]


def is_liquid(chain):
    return chain in ["liquidv1", "liquidtestnet", "elementsregtest", "elreg"]


def normalize_address(addr: str) -> str:
    """
    Normalized address to canonical form.
    Upper bech32 -> lower bech32, but doesn't touch base58 addresses.
    """
    # bitcoin hrps
    hrps = {net["bech32"] + "1" for net in NETWORKS.values()}
    # liquid hrps
    # Blech32 addresses are intended for confidential assets
    hrps = hrps.union(
        {net["blech32"] + "1" for net in NETWORKS.values() if "blech32" in net}
    )
    if addr.lower().startswith(tuple(hrps)):
        return addr.lower()
    return addr


def locked(customlock=defaultlock):
    """
    @locked(lock) decorator.
    Make sure you are not calling
    @locked function from another @locked function
    with the same lock argument.
    """

    def wrapper(fn):
        def wrapper_fn(*args, **kwargs):
            with customlock:
                return fn(*args, **kwargs)

        return wrapper_fn

    return wrapper


def to_ascii20(name: str) -> str:
    """
    Converts the name to max 20 ascii-only characters.
    """
    # ascii characters have codes from 0 to 127
    # but 127 is "delete" and we don't want it
    return "".join([c for c in name if ord(c) < 127])[:20]


def alias(name):
    """
    Create a filesystem-friendly alias from a string.
    Replaces space with _ and keeps only alphanumeric chars.
    """
    name = name.replace(" ", "_")
    return "".join(x for x in name if x.isalnum() or x == "_").lower()


def deep_update(d, u):
    """updates the dict d with the dict u"""
    for k, v in six.iteritems(u):
        dv = d.get(k, {})
        if not isinstance(dv, collections.abc.Mapping):
            d[k] = v
        elif isinstance(v, collections.abc.Mapping):
            d[k] = deep_update(dv, v)
        else:
            d[k] = v
    return d


def load_jsons(folder, key=None):
    # get all json files (not hidden)
    files = [
        f for f in os.listdir(folder) if f.endswith(".json") and not f.startswith(".")
    ]
    files.sort(key=lambda x: os.path.getmtime(os.path.join(folder, x)))
    dd = OrderedDict()
    for fname in files:
        try:
            d = read_json_file(os.path.join(folder, fname))
            if key is None:
                dd[fname[:-5]] = d
            else:
                d["fullpath"] = os.path.join(folder, fname)
                d["alias"] = fname[:-5]
                dd[d[key]] = d
        except Exception as e:
            logger.exception(e)
    return dd


def set_loglevel(app, loglevel_string):
    logger.info(
        "Setting Loglevel to %s (Check the next log-line(s) whether it's effective here)"
        % loglevel_string
    )
    loglevels = {"WARN": logging.WARN, "INFO": logging.INFO, "DEBUG": logging.DEBUG}
    logging.getLogger().setLevel(loglevels[loglevel_string])
    logger.warning("Loglevel-Test: This is a warn-message!")
    logger.info("Loglevel-Test: This is an info-message!")
    logger.debug("Loglevel-Test: This is an debug-message!")


def get_loglevel(app):
    loglevels = {logging.WARN: "WARN", logging.INFO: "INFO", logging.DEBUG: "DEBUG"}
    return loglevels[app.logger.getEffectiveLevel()]


def hwi_get_config(specter):
    config = {"whitelisted_domains": "http://127.0.0.1:25441/"}
    # if hwi_bridge_config.json file exists - load from it
    if os.path.isfile(os.path.join(specter.data_folder, "hwi_bridge_config.json")):
        fname = os.path.join(specter.data_folder, "hwi_bridge_config.json")
        file_config = read_json_file(fname)
        deep_update(config, file_config)
    # otherwise - create one and assign unique id
    else:
        save_hwi_bridge_config(specter, config)
    return config


def save_hwi_bridge_config(specter, config):
    if "whitelisted_domains" in config:
        whitelisted_domains = ""
        for url in config["whitelisted_domains"].split():
            if not url.endswith("/") and url != "*":
                # make sure the url end with a "/"
                url += "/"
            whitelisted_domains += url.strip() + "\n"
        config["whitelisted_domains"] = whitelisted_domains
    fname = os.path.join(specter.data_folder, "hwi_bridge_config.json")
    write_json_file(config, fname)


def der_to_bytes(derivation):
    items = derivation.split("/")
    if len(items) == 0:
        return b""
    if items[0] == "m":
        items = items[1:]
    if len(items) > 0 and items[-1] == "":
        items = items[:-1]
    res = b""
    for item in items:
        index = 0
        if item[-1] == "h" or item[-1] == "'":
            index += 0x80000000
            item = item[:-1]
        index += int(item)
        res += index.to_bytes(4, "little")
    return res


def get_devices_with_keys_by_type(app, cosigners, wallet_type):
    devices = []
    prefix = "tpub"
    if not is_testnet(app.specter.chain):
        prefix = "xpub"
    for cosigner in cosigners:
        device = copy.deepcopy(cosigner)
        allowed_types = ["", wallet_type]
        if wallet_type == "simple":
            allowed_types += ["sh-wpkh", "wpkh", "tr"]
        elif wallet_type == "multisig":
            allowed_types += ["sh-wsh", "wsh"]
        device.keys = sorted(
            [
                key
                for key in device.keys
                if key.xpub.startswith(prefix)
                and (key.key_type in allowed_types or wallet_type == "*")
            ],
            key=lambda k: k.original == k.xpub,
        )
        devices.append(device)
    return devices


def clean_psbt(b64psbt):
    try:
        psbt = PSBT.from_string(b64psbt)
    except:
        psbt = PSET.from_string(b64psbt)
    for inp in psbt.inputs:
        if inp.witness_utxo is not None and inp.non_witness_utxo is not None:
            inp.non_witness_utxo = None
    return psbt.to_string()


def bcur2base64(encoded):
    raw = bcur_decode(encoded.split("/")[-1])
    return binascii.b2a_base64(raw).strip()


def get_txid(tx):
    try:
        t = Transaction.from_string(tx)
    except:
        t = LTransaction.from_string(tx)
    for inp in t.vin:
        inp.scriptSig = Script(b"")
    return t.txid().hex()


def get_startblock_by_chain(specter):
    if specter.info["chain"] == "main":
        if not specter.info["pruned"] or specter.info["pruneheight"] < 481824:
            startblock = 481824
        else:
            startblock = specter.info["pruneheight"]
    else:
        if not specter.info["pruned"]:
            startblock = 0
        else:
            startblock = specter.info["pruneheight"]
    return startblock


def notify_upgrade(app, flash):
    """If a new version is available, notifies the user via flash
    that there is an upgrade to specter.desktop
    :return the current version
    """
    if app.specter.version.upgrade:
        flash(
            f"Upgrade notification: new version {app.specter.version.latest} is available.",
            "info",
        )
    return app.specter.version.current


def is_ip_private(ip):
    # https://en.wikipedia.org/wiki/Private_network
    priv_lo = re.compile(r"^127\.\d{1,3}\.\d{1,3}\.\d{1,3}$")
    priv_24 = re.compile(r"^10\.\d{1,3}\.\d{1,3}\.\d{1,3}$")
    priv_20 = re.compile(r"^192\.168\.\d{1,3}.\d{1,3}$")
    priv_16 = re.compile(r"^172.(1[6-9]|2[0-9]|3[0-1]).[0-9]{1,3}.[0-9]{1,3}$")

    res = (
        ip == "localhost"
        or priv_lo.match(ip)
        or priv_24.match(ip)
        or priv_20.match(ip)
        or priv_16.match(ip)
    )
    return res is not None


def get_address_from_dict(data_dict):
    # TODO: Remove this helper function in favor of simple ["address"]
    # when support for Bitcoin Core version < 22 is dropped
    addr = data_dict.get("addresses")
    if not addr:
        addr = data_dict.get("scriptPubKey", {}).get("addresses")
    if addr:
        addr = addr[0]
    if not addr:
        addr = data_dict.get("address")
    if addr and addr != "Fee":
        return addr
    raise RuntimeError(f"Missing address info in object {data_dict}")



from binascii import a2b_base64
from typing import List
from . import DeviceTypes
from ..util.base43 import b43_encode
from ..device import Device


class Electrum(Device):
    device_type = DeviceTypes.ELECTRUM
    name = "Electrum"
    icon = "electrum_icon.svg"
    template = "device/new_device/new_device_keys_electrum.jinja"

    sd_card_support = True
    qr_code_support = True
    qr_code_animate = "off"

    def create_psbts(self, base64_psbt, wallet):
        # remove non_witness utxo for QR code
        updated_psbt = wallet.fill_psbt(base64_psbt, non_witness=False, xpubs=False)
        psbts = {"qrcode": b43_encode(a2b_base64(updated_psbt)), "sdcard": base64_psbt}
        return psbts
